# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: sinogram_alignment
   :platform: Unix
   :synopsis: A plugin to determine the centre of rotation of a sinogram and\
       to align the rows of a sinogram e.g. in the case of motor backlash.

.. moduleauthor:: Stephen Price

"""

import logging
from scipy import ndimage
from scipy.optimize import curve_fit
import numpy as np

from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter
from savu.data.plugin_list import CitationInformation
from savu.plugins.driver.cpu_plugin import CpuPlugin


@register_plugin
class SinogramAlignment(BaseFilter, CpuPlugin):
    """
    The centre of mass of each row is determined and then a sine function fit
    through these to determine the centre of rotation.  The residual between\
    each centre of mass and the sine function is then used to align each row.

    :param threshold: e.g. a.b will set all values above a to b. Default: None.
    :param p0: Initial guess for the parameters of \
        scipy.optimize.curve_fit. Default: (1, 1, 1).
    :param type: Either centre_of_mass or shift, with the latter requiring\
        ProjectionVerticalAlignment prior to this \
        plugin. Default: 'centre_of_mass'.
    """

    def __init__(self):
        logging.debug("initialising Sinogram Alignment")
        super(SinogramAlignment,
              self).__init__("SinogramAlignment")
        self.com_y = None

    def pre_process(self):
        if self.parameters['type'] == 'shift':
            self.com_y = self.get_in_datasets()[0].meta_data.get(
                'proj_align_shift')[:, 1]
        self.com_x = \
            self.get_in_datasets()[0].meta_data.get('rotation_angle')
        data = self.get_in_datasets()[0]
        self.sl = [slice(None)]*len(data.get_shape())
        self.slice_dir = self.get_plugin_in_datasets()[0].get_slice_dimension()

#    def filter_frames(self, data):
#        """
#        Should be overloaded by filter classes extending this one
#
#        :param data: The data to filter
#        :type data: ndarray
#        :returns:  The filtered image
#        """
#        nFrames = data[0].shape[self.slice_dir]
#        result = np.empty_like(data[0])
#        for i in range(nFrames):
#            self.sl[self.slice_dir] = i
#            sino = data[0][self.sl]
#            if self.parameters['threshold']:
#                a, b = self.parameters['threshold'].split('.')
#                sino[sino > a] = b
#            com_y = self.com_y if self.com_y is not None else self._com_y(sino)
#            shifted = self._shift(sino, self.com_x, com_y)
#            result[self.sl] = \
#                shifted.reshape(shifted.shape[0], shifted.shape[1])
#        return data

    def process_frames(self, data):
        """
        Should be overloaded by filter classes extending this one

        :param data: The data to filter
        :type data: ndarray
        :returns:  The filtered image
        """
        nFrames = data[0].shape[self.slice_dir]
        result = np.empty_like(data[0])
        for i in range(nFrames):
            self.sl[self.slice_dir] = i
            sino = data[0][tuple(self.sl)]
            if self.parameters['threshold']:
                a, b = self.parameters['threshold'].split('.')
                sino[sino > a] = b
            com_y = self.com_y if self.com_y is not None else self._com_y(sino)
            shifted = self._shift(sino, self.com_x, com_y)
            result[tuple(self.sl)] = shifted.reshape(shifted.shape[0], shifted.shape[1])
        return result

    def _sinfunc(self, data, a, b, c):
        return (a*np.sin(np.deg2rad(data-b)))+c

    def _shift(self, sinogram, com_x, com_y):
        fitpars, covmat = \
            curve_fit(self._sinfunc, com_x, com_y, p0=self.parameters['p0'])
        variances = covmat.diagonal()
        std_devs = np.sqrt(variances)
        #residual = com_y - self._sinfunc(com_x, *fitpars)
        residual = self._sinfunc(com_x, *fitpars) - com_y
        centre_of_rotation_shift = residual
        np.array_split(sinogram, sinogram.shape[0], axis=0)
        n = 0
        shifted_sinogram = []
        for row in sinogram:
            shifted_sinogram_row = \
                ndimage.interpolation.shift(row, [centre_of_rotation_shift[n]],
                                            mode='nearest')
            shifted_sinogram.append(shifted_sinogram_row)
            n += 1
            output = np.vstack(shifted_sinogram)
        return output

    def _com_y(self, sinogram):
        com_y = []
        for row in sinogram:
            com = ndimage.measurements.center_of_mass(row)
            com_y.append(com[0])
        return com_y

    def get_plugin_pattern(self):
        return 'SINOGRAM'

    def get_max_frames(self):
        return 'multiple'

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The Tomographic filtering performed in this processing " +
             "chain is derived from this work.")
        cite_info.bibtex = \
            ("@Article{C4CP04488F, \n" +
             "author ={Price, S. W. T. and Ignatyev, K. and Geraki, K. and \
             Basham, M. and Filik, J. and Vo, N. T. and Witte, P. T. and \
             Beale, A. M. and Mosselmans, J. F. W.}, \n" +
             "title  ={Chemical imaging of single catalyst particles with \
             scanning [small mu ]-XANES-CT and [small mu ]-XRF-CT}, \n" +
             "journal  ={Phys. Chem. Chem. Phys.}, \n" +
             "year  ={2015}, \n" +
             "volume  ={17}, \n" +
             "issue  ={1}, \n" +
             "pages  ={521-529}, \n" +
             "publisher  ={The Royal Society of Chemistry}, \n" +
             "doi  ={10.1039/C4CP04488F}, \n" +
             "url  ={https://doi.org/10.1039/C4CP04488F}, \n" +
             "}")
        cite_info.endnote = \
            ("%0 Journal Article\n" +
             "%T Chemical imaging of single catalyst particles with scanning \
            [small mu ]-XANES-CT and [small mu ]-XRF-CT\n" +
             "%A Price, Stephen W.T.\n" +
             "%A Ignatyev, Konstantin\n" +
             "%A Geraki, Kalotina\n" +
             "%A Basham, Mark\n" +
             "%A Filik, Jacob\n" +
             "%A Vo, Nghia T.\n" +
             "%A Witte, Peter T.\n" +
             "%A Beale, Andrew M.\n" +
             "%A Mosselmans, J. Fred W.\n" +
             "%J Physical Chemistry Chemical Physics\n" +
             "%V 17\n" +
             "%N 1\n" +
             "%P 521-529\n" +
             "%@ 1094-4087\n" +
             "%D 2015\n" +
             "%I Royal Society of Chemistry")
        cite_info.doi = "doi: 10.1039/c4cp04488f"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: iterative_plugin
   :platform: Unix
   :synopsis: Driver for iterative plugins

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.driver.plugin_driver import PluginDriver


class IterativePlugin(PluginDriver):
    """
    Allows the plugin to be repeated, keeping the parameters from the previous
    iteration.
    """

    def __init__(self):
        super(IterativePlugin, self).__init__()
        self._ip_iteration = 0
        self._ip_fixed_iterations = False
        self._ip_complete = False
        self._ip_data_dict = {}
        self._ip_pattern_dict = {}
        self._ip_data_dict['iterating'] = {}
        self._ip_pattern_dict['iterating'] = {}

    def _run_plugin(self, exp, transport):
        """ Runs the pre_process, process and post_process methods.
        """
        self.__set_original_datasets()

        while not self._ip_complete:
            print ("Iteration", self._ip_iteration, "...")
            self.__set_datasets()  # change the pattern in this function?
            self._run_plugin_instances(transport, self.get_communicator())
            if transport.no_processing:
                self.set_processing_complete()
            if self._ip_fixed_iterations and \
                    self._ip_iteration == self._ip_fixed_iterations-1:
                self.set_processing_complete()
            self._ip_iteration += 1
        self.__finalise_datasets()

    def __set_datasets(self):
        params = self.parameters
        if self._ip_iteration in list(self._ip_data_dict.keys()):
            params['in_datasets'] = self._ip_data_dict[self._ip_iteration][0]
            params['out_datasets'] = self._ip_data_dict[self._ip_iteration][1]
        elif self._ip_iteration > 0:
            p = [params['in_datasets'], params['out_datasets']]
            for s1, s2 in self._ip_data_dict['iterating'].items():
                a = [0, p[0].index(s1)] if s1 in p[0] else [1, p[1].index(s1)]
                b = [0, p[0].index(s2)] if s2 in p[0] else [1, p[1].index(s2)]
                p[a[0]][a[1]], p[b[0]][b[1]] = p[b[0]][b[1]], p[a[0]][a[1]]

#    def __set_patterns(self):
#        # The pattern dict key
#        # pattern_dict can be:
#        # dict[int=itr] = [list of string patterns]
#        # dict[int=itr] = {dictionary of <dataset, pattern> key value pairs}
#        # dict['iterating'] = [list of string patterns]
#        # dict['iterating'] = {dictionary of <dataset, pattern> key value pairs}
#        params = self.parameters
#        if self._ip_iteration in self._ip_pattern_dict.keys():
#            # initial attempt at recalculating every time
#            params['in_datasets'] =


    def get_iteration(self):
        return self._ip_iteration

    def __finalise_datasets(self):
        for s1, s2 in self._ip_data_dict['iterating'].items():
            name = s1.get_name()
            name = name if 'itr_clone' not in name else s2.get_name()
            final_dataset = s1 if s1 in self.parameters['out_datasets'] else s2
            obsolete = s1 if s1 is not final_dataset else s2
            obsolete.remove = True
            # switch names if necessary
            if final_dataset.get_name() != name:
                temp = obsolete
                self.exp.index['out_data'][name] = final_dataset
                self.exp.index['out_data'][s2.get_name()] = temp

    def set_processing_complete(self):
        self._ip_complete = True

    def set_iterations(self, nIterations):
        if isinstance(nIterations, int):
            self._ip_fixed_iterations = nIterations
        else:
            raise Exception('nIterations should be an integer.')

    def set_iteration_datasets(self, itr, in_data, out_data, pattern=None):
        self._ip_data_dict[itr] = [in_data, out_data]
        self._ip_pattern_dict[itr] = pattern

    def create_clone(self, clone, data):
        clone.create_dataset(data)
        clone.data_info.set('clone', data.get_name())
        # alternate a dataset with its clone
        self.set_alternating_datasets(data, clone)

    def set_alternating_datasets(self, d1, d2):
        names = [d1.get_name(), d2.get_name()]
        if not any([True if 'itr_clone' in i else False for i in names]):
            raise Exception('Alternating datasets must contain a clone.  These'
                            ' are found at the end of the out_datasets list')
        self._ip_data_dict['iterating'][d1] = d2

    def set_alternating_patterns(self, patterns):
        if isinstance(patterns, list):
            self._ip_pattern_dict['iterating'] = patterns
        elif isinstance(patterns, dict):
            for data, plist in patterns:
                self._ip_pattern_dict['iterating'][data] = plist
        else:
            raise Exception("Please pass the patterns as a list or a dict.")

    def get_alternating_datasets(self):
        return self._ip_data_dict['iterating']

    def __set_original_datasets(self):
        self.in_data = self.parameters['in_datasets']
        self.out_data = self.parameters['out_datasets']

    def get_original_datasets(self):
        return self.in_data, self.out_data



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: sinogram_clean
   :platform: Unix
   :synopsis: A plugin to find the center of rotation per frame

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
from savu.plugins.driver.cpu_plugin import CpuPlugin

import math
import numpy as np
import pyfftw.interfaces.scipy_fftpack as fft

from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter


@register_plugin
class SinogramClean(BaseFilter, CpuPlugin):
    """
    A plugin to calculate the centre of rotation using the Vo Method

    :param ratio: The ratio between the size of object and FOV of \
        the camera. Default: 2.0.
    :param row_drop: Drop lines around vertical center of the \
        mask. Default: 20.
    :param out_datasets: The default names. Default: [].
    """

    def __init__(self):
        super(SinogramClean, self).__init__("SinogramClean")

    def _create_mask(self, Nrow, Ncol, obj_radius):
        du = 1.0 / Ncol
        dv = (Nrow-1.0)/(Nrow*2.0*math.pi)
        cen_row = np.ceil(Nrow / 2.0)-1
        cen_col = np.ceil(Ncol / 2.0)-1
        drop = self.parameters['row_drop']
        mask = np.zeros((Nrow, Ncol), dtype=np.float32)
        for i in range(Nrow):
            num1 = np.round(((i-cen_row)*dv/obj_radius)/du)
            (p1, p2) = np.clip(np.sort((-num1+cen_col, num1+cen_col)), 0, Ncol-1)
            mask[i, p1:p2+1] = np.ones(p2-p1+1, dtype=np.float32)
        if drop < cen_row:
            mask[cen_row-drop:cen_row+drop+1, :] = np.zeros((2*drop + 1, Ncol), dtype=np.float32)
        return mask

    def process_frames(self, data):
        sino = data[0]
        sino2 = np.fliplr(sino[1:])
        (Nrow, Ncol) = sino.shape
        mask = self._create_mask(
            2*Nrow-1, Ncol, 0.5*self.parameters['ratio']*Ncol)

        FT1 = fft.fftshift(fft.fft2(np.vstack((sino, sino2))))
        sino = fft.ifft2(fft.ifftshift(FT1 - FT1*mask))
        return sino[0:Nrow].real

    def get_max_frames(self):
        return 1



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 .. module:: module to threshold the data (less, lessequal, equal, greater, greaterequal) than the given value
   :platform: Unix
   :synopsis: module to threshold the data (less, lessequal, equal, greater, greaterequal) than the given value

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""


from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import numpy as np

@register_plugin
class DataThreshold(Plugin, CpuPlugin):
    """
    The module to threshold the data (less, lessequal, equal, greater, greaterequal) than the given value,\
    based on the condition the data values will be replaced by the provided new value

    :param inequality_condition: set to less, lessequal, equal, greater, greaterequal. Default: 'greater'.
    :param given_value: The value to be replaced with by inequality_condition. Default: 1.   
    :param new_value: The value to be replaced with by inequality_condition. Default: 1.   
    """

    def __init__(self):
        super(DataThreshold, self).__init__("DataThreshold")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('VOLUME_XZ', 'single')        
        out_dataset[0].create_dataset(in_dataset[0])
        out_pData[0].plugin_data_setup('VOLUME_XZ', 'single')

    def process_frames(self, data):
        if (self.parameters['inequality_condition'] == 'less'):
            indeces = np.where(data[0] < self.parameters['given_value'])
        elif (self.parameters['inequality_condition'] == 'lessequal'):
            indeces = np.where(data[0] <= self.parameters['given_value'])
        elif (self.parameters['inequality_condition'] == 'greater'):
            indeces = np.where(data[0] > self.parameters['given_value'])
        elif (self.parameters['inequality_condition'] == 'greaterequal'):
            indeces = np.where(data[0] >= self.parameters['given_value'])
        else:
            indeces = np.where(data[0] == self.parameters['given_value'])                
        corr_data = np.copy(data[0])
        corr_data[indeces] = self.parameters['new_value']
        return corr_data

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1

    def get_max_frames(self):
        return 'single'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: mc_near_absorption_correction
   :platform: Unix
   :synopsis: A plugin apply hogans xrf absorption correction using stxm data

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""

import numpy as np
import logging
from savu.plugins.absorption_corrections.base_absorption_correction \
    import BaseAbsorptionCorrection
from savu.plugins.utils import register_plugin


@register_plugin
class McNearAbsorptionCorrection(BaseAbsorptionCorrection):
    """
    McNears absorption correction, takes in a normalised absorption sinogram \
    and xrf sinogram stack.

    :param in_datasets: A list of the dataset(s) to \
        process. Default: ['xrf','stxm'].
    """

    def __init__(self):
        logging.debug("Starting McNear Absorption correction")
        super(McNearAbsorptionCorrection,
              self).__init__("McNearAbsorptionCorrection")

    def pre_process(self):
        compound = self.parameters['compound']
        density = self.parameters['density']
        mData = self.get_in_meta_data()[0]

        mono_energy = mData.get('mono_energy')
        try:
            peak_energy = mData.get('PeakEnergy')
        except KeyError:
            logging.debug('No PeakEnergy: trying with the fullSpectrum')
            try:
                in_dataset, out_datasets = self.get_datasets()
                in_dataset[0].get_data_patterns()['SPECTRUM']
                peak_energy = list(mData.get('energy'))
            except KeyError:
                logging.debug("No PeakEnergy or energy axis. This won't work")
                raise

        pump_mu = self.get_mu(compound, float(mono_energy), density)
        peak_mu = self.get_mu(compound, list(peak_energy), density)
        self.atten_ratio: float = [pm / pump_mu for pm in peak_mu]

        theta = mData.get('rotation_angle')
        self.dtheta = theta[1]-theta[0]
        logging.debug('The rotation step is %s' % str(self.dtheta))
        if np.abs(self.dtheta)>10.0:
            logging.warning('The theta step is greater than 10 degrees! Watch out!')
        self.npix_displacement = self.parameters['azimuthal_offset']//self.dtheta
        logging.debug('This gives a pixel offset of %s' % str(self.npix_displacement))

    def process_frames(self, data):
        xrf = data[0]
        stxm_orig = data[1]
        logging.debug('the xrf shape is %s' % str(xrf.shape))
        logging.debug('the stxm shape is %s' % str(stxm_orig.shape))
        # take the log here, we assume it is monitor corrected already
        stxm = stxm_orig
        # now correct for the rotation offset
        absorption = np.roll(stxm,int(self.npix_displacement), axis=0)
        num_channels = self.get_num_channels()
        corrected_xrf = np.zeros_like(xrf)
        for i in range(num_channels):
            fluo_sino = xrf[:,:,i]
            corrected_xrf[:,:,i], corr_fac = self.correct_sino(self.atten_ratio[i], fluo_sino, absorption)
            logging.debug('For channel %s, min correction: %s, max correction: %s' % (str(i),
                                                                                      str(np.min(corr_fac)),
                                                                                      str(np.max(corr_fac))))
        return corrected_xrf

    def correct_sino(self,Ti_ratio, FFI0_Ti, absorption):
        trans_ave_array = np.ndarray(shape=FFI0_Ti.shape, dtype=np.float64)
        for n in range(len(absorption)): #columns, axis=0
            row = absorption[n]
            trans_ave_array[n] = row
            for m in range(1, len(row), 1): #row, axis=1
                trans_ave_array[n][m] = np.average(row[0:m]) #average of all absorption between pixel 0 and pixel - removes t so should be mu only

        trans_ave_array = np.nan_to_num(trans_ave_array)
        exponent_Ti = self.get_exponent_Ti_mu(Ti_ratio, absorption, trans_ave_array)
        FFI0_corrected_Ti = np.multiply(FFI0_Ti, exponent_Ti)
        return FFI0_corrected_Ti, exponent_Ti

    def get_exponent_Ti_mu(self, Ti_ratio, absorption, trans_ave_array):
        FF_Ti_xray_mu = np.multiply(trans_ave_array, Ti_ratio)
        exponent_Ti = np.add(FF_Ti_xray_mu, absorption)
        exponent_Ti = np.exp(exponent_Ti)
        return exponent_Ti



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Convert 360-180 sinogram.
   :platform: Unix
   :synopsis: A plugin working in sinogram space to convert 0-360 degree\\
    sinogram to 0-180 sinogram.
.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.data.plugin_list import CitationInformation
import numpy as np
import scipy.ndimage as ndi


@register_plugin
class Convert360180Sinogram(Plugin, CpuPlugin):
    """

    Method to convert the 0-360 degree sinogram to 0-180 sinogram.
    :param center: Center of rotation. Default: 0.0
    :param out_datasets: Create a list of the dataset(s) to \
        create. Default: ['in_datasets[0]', 'cor'].

    """

    def __init__(self):
        super(Convert360180Sinogram, self).__init__(
            "Convert360180Sinogram")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()

        self.center = 0.0
        key = "centre_of_rotation"
        if key in list(in_dataset[0].meta_data.get_dictionary().keys()):
            self.center = in_dataset[0].meta_data.get(key)

        old_shape = in_dataset[0].get_shape()
        width_dim = in_dataset[0].get_data_dimension_by_axis_label(
            'detector_x')
        self.height1_dim = in_dataset[0].get_data_dimension_by_axis_label(
            'detector_y')

        height_dim = in_dataset[0].get_data_dimension_by_axis_label(
            'rotation_angle')
        new_shape = list(old_shape)
        new_shape[width_dim] *= 2
        new_shape[height_dim] = np.int16(np.ceil(new_shape[height_dim] / 2.0))

        out_dataset[0].create_dataset(patterns=in_dataset[0],
                                      axis_labels=in_dataset[0],
                                      shape=tuple(new_shape))
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')

        slice_dirs = np.array(in_dataset[0].get_slice_dimensions())
        new_shape = (np.prod(np.array(old_shape)[slice_dirs]), 1)
        self.orig_shape = (np.prod(np.array(old_shape)[slice_dirs]), 1)

        out_dataset[1].create_dataset(shape=new_shape,
                                      axis_labels=['x.pixels', 'y.pixels'],
                                      transport='hdf5')
        out_dataset[1].add_pattern("METADATA", core_dims=(1,), slice_dims=(0,))

        out_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[1].plugin_data_setup('METADATA', 'single')

    def pre_process(self):
        out_dataset = self.get_out_datasets()[0]
        in_pData = self.get_plugin_in_datasets()
        width_dim = in_pData[0].get_data_dimension_by_axis_label('detector_x')
        height_dim = in_pData[0].get_data_dimension_by_axis_label(
            'rotation_angle')
        sino_shape = list(in_pData[0].get_shape())
        self.width = sino_shape[width_dim]
        self.height = sino_shape[height_dim]
        center_manu = float(self.parameters['center'])
        if center_manu != 0.0:
            self.center = center_manu
        self.mid_width = self.width / 2.0
        if (self.center <= 0) or (self.center > self.width):
            self.center = self.mid_width
        center_int = np.int16(np.floor(self.center))
        self.subpixel_shift = self.center - center_int
        if self.center < self.mid_width:
            self.overlap = 2 * center_int
            self.cor = self.width + center_int
        else:
            self.overlap = 2 * (self.width - center_int)
            self.cor = center_int

        self.new_height = np.int16(np.ceil(self.height / 2.0))
        list_angle = out_dataset.meta_data.get("rotation_angle")
        list_angle = list_angle[0:self.new_height]

        out_dataset.meta_data.set("rotation_angle", list_angle)
        list_wedge = np.linspace(1.0, 0.0, self.overlap)
        self.mat_wedge_left = \
            np.ones((self.new_height, self.width), dtype=np.float32)
        self.mat_wedge_left[:, -self.overlap:] = np.float32(list_wedge)
        self.mat_wedge_right = np.fliplr(self.mat_wedge_left)

    def process_frames(self, data):
        sinogram = np.copy(data[0])
        sinogram = ndi.interpolation.shift(sinogram, (0, -self.subpixel_shift),
                                           prefilter=False, mode='nearest')
        sinogram1 = sinogram[:self.new_height]
        sinogram2 = np.fliplr(sinogram[-self.new_height:])
        sinocombine = np.zeros((self.new_height, 2 * self.width),
                               dtype=np.float32)
        if self.center < self.mid_width:
            num1 = np.mean(np.abs(sinogram1[:, :self.overlap]))
            num2 = np.mean(np.abs(sinogram2[:, -self.overlap:]))
            sinogram2 = sinogram2 * num1 / num2
            sinogram1 = sinogram1 * self.mat_wedge_right
            sinogram2 = sinogram2 * self.mat_wedge_left
            sinocombine[:, 0:self.overlap] = sinogram2[:, 0:1]
            sinocombine[:, self.overlap:self.overlap + self.width] = sinogram2
            sinocombine[:, -self.width:] += sinogram1
        else:
            num1 = np.mean(np.abs(sinogram1[:, -self.overlap:]))
            num2 = np.mean(np.abs(sinogram2[:, :self.overlap]))
            sinogram2 = sinogram2 * num1 / num2
            sinogram1 = sinogram1 * self.mat_wedge_left
            sinogram2 = sinogram2 * self.mat_wedge_right
            sinocombine[:, 0:self.width] = sinogram1
            sinocombine[:, self.width - self.overlap:
                        2 * self.width - self.overlap] += sinogram2
            sinocombine[:, -self.overlap:] = sinogram2[:, -1:]
        return [sinocombine, np.array([self.cor])]

    def nOutput_datasets(self):
        return 2



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_ptycho
   :platform: Unix
   :synopsis: A base class for all ptychographic analysis methods

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
import logging
import numpy as np


class BasePtycho(Plugin, CpuPlugin):  # also make one for gpu
    """
    A base plugin for doing ptychography. Other ptychography plugins should \
    inherit from this.
    :param in_datasets: A list of the dataset(s) to process. Default: [].
    :param out_datasets: A list of the dataset(s) to \
        process. Default: ['probe', 'object_transmission', 'positions'].
    """

    def __init__(self, name):
        super(BasePtycho, self).__init__(name)

    def setup(self):
        self.exp.log(self.name + " Setting up the ptycho")
        in_dataset, out_dataset = self.get_datasets()

        in_meta_data = in_dataset[0].meta_data# grab the positions from the metadata
        logging.debug('getting the positions...')
        self.positions = in_meta_data.get('xy') # get the positions and bind them

        # lets set up the axis labels for output datasets
        position_labels, probe_labels, object_labels, self.sh = self.setup_axis_labels(in_dataset)
#        print "probe labels are:"+str(probe_labels)
#        print "object labels are:"+str(object_labels)
#        print "position labels are:"+str(position_labels)
        # Now create the datasets and work out the patterns
        ### PROBE ###
        probe = out_dataset[0]
#         probe_shape = in_dataset[0].get_shape()[-2:] + (self.get_num_probe_modes(),)

        self.set_size_probe(in_dataset[0].get_shape()[-2:])
        logging.debug("##### PROBE #####")
        #print("probe shape is:%s",str(self.get_size_probe()))
        probe.create_dataset(axis_labels=probe_labels,
                            shape=self.get_size_probe()) # create the dataset
        self.probe_pattern_setup(probe_labels, probe)

        ### OBJECT ####
        self.set_size_object(in_dataset[0], self.get_positions(),
                             self.get_pixel_size())
        object_trans = out_dataset[1]
        object_shape = self.sh + self.get_size_object()
        logging.debug("##### OBJECT #####")
        #print("object shape is:%s",str(object_shape))
#         print object_labels

        object_trans.create_dataset(axis_labels=object_labels,
                                    shape=object_shape) # create the dataset

        self.object_pattern_setup(object_labels, object_trans)

        ### POSITIONS ###
        logging.debug('##### POSITIONS #####')
        positions = out_dataset[2]
        #print self.sh, self.get_positions().shape
        positions_shape = self.sh + self.get_positions().shape[-2:]
        logging.debug('positions shape is:%s',str(positions_shape))
        #print "positions shape",positions_shape
        positions.create_dataset(axis_labels=position_labels,
                                 shape=positions_shape)

        rest_pos = list(range(len(position_labels)))

        pos_md = \
            {'core_dims':tuple(set(rest_pos) - set([0])), 'slice_dims':(0,)}
        positions.add_pattern("CHANNEL", **pos_md)
        '''
        now we need to tell the setup what we want as input shapes, output shapes, and the number of each of them in one go.
        '''
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup(self.get_plugin_pattern(), self.get_max_frames())
        out_pData[0].plugin_data_setup("PROJECTION", self.get_num_probe_modes())
        out_pData[1].plugin_data_setup("PROJECTION", self.get_num_object_modes())
        out_pData[2].plugin_data_setup("CHANNEL", self.get_max_frames())
        self.exp.log(self.name + " End")

    '''
    The below methods influence the set-up and can be over-ridden depending on which software package we are using

    '''

    def get_plugin_pattern(self):
        '''
        sets the pattern to work in. In this case we consider a ptycho scan to be a 4D_SCAN.
        '''
        return "4D_SCAN"

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 3

    def get_num_probe_modes(self):
        return 1

    def get_num_object_modes(self):
        return 1

    def get_positions(self):
        return self.positions

    def get_pixel_size(self):
        return 30e-9

    def set_size_object(self, dataset,positions, pobj=33e-9):
        '''
        returns tuple
        '''
#         print "positions is "+str(self.get_positions().shape)
        x,y = self.get_positions()[0],self.get_positions()[1]
        probe_size = self.get_size_probe()
        x_fov = np.max(x)-np.min(x)
        y_fov = np.max(y)-np.min(y)
        xsize = int(x_fov//pobj) + probe_size[0]
        ysize = int(y_fov//pobj) + probe_size[1]
        self.obj_shape = xsize,ysize,self.get_num_object_modes()

    def get_size_object(self):
        return self.obj_shape

    def set_size_probe(self,val):
        self.probe_size = (1,)+val + (self.get_num_probe_modes(),)

    def get_size_probe(self):
        '''
        returns tuple
        '''
        return self.probe_size

    def get_max_frames(self):
        return 'single'

    def get_output_axis_units(self):
        return 'nm'

    def probe_pattern_setup(self, probe_labels, probe):
        '''
        This is where we set up the patterns, we need to add, PROJECTIONS, SINOGRAMS, TIMESERIES and SPECTRA
        I've created the TIMESERIES because we could in theory have a time series of spectra
        probe_patterns: PROJECTION, TIMESERIES (for each projection), SPECTRUM (for each energy)
        object_patterns: PROJECTION, SINOGRAM, SPECTRUM (for each energy)
        position_patterns: 1D_METADATA

        '''
        probe_dims = len(probe_labels) # the number of dimensions from the axis labels
        rest_probe = list(range(probe_dims)) # all the dimensions we have
        self.set_projection_pattern(probe, rest_probe)
        self.set_probe_rotation_patterns(probe, rest_probe)
        self.set_probe_energy_patterns(probe, rest_probe)

    def object_pattern_setup(self, object_labels, object_trans):
        '''
        This is where we set up the patterns, we need to add, PROJECTIONS, SINOGRAMS, TIMESERIES and SPECTRA
        I've created the TIMESERIES because we could in theory have a time series of spectra
        probe_patterns: PROJECTION, TIMESERIES (for each projection), SPECTRUM (for each energy)
        object_patterns: PROJECTION, SINOGRAM, SPECTRUM (for each energy)
        position_patterns: 1D_METADATA

        '''
        obj_dims = len(object_labels) # the number of dimensions from the axis labels
#         print "object has "+str(obj_dims)+"dimensions"
        rest_obj = list(range(obj_dims)) # all the dimensions we have
        self.set_projection_pattern(object_trans, rest_obj)
        self.set_object_rotation_patterns(object_trans, rest_obj)
        self.set_object_energy_patterns(object_trans, rest_obj)

    def setup_axis_labels(self, in_dataset):
        '''
        This is where we set up the axis labels
        the 4D scan will contain labels that are: 'xy', 'detectorX', 'detectorY', but the data
        itself may be scanned in energy or rotation or something else. We want to remove all the above,
        and amend them to be the following (preferably with additional scan axes at the front):
        probe: 'x','y','mode_idx'
        object: 'x','y','mode_idx'
        positions: 'xy'
        '''
        PATTERN_LABELS = ['xy', 'detectorX', 'detectorY']
        in_labels = in_dataset[0].data_info.get('axis_labels') # this is a list of dictionarys
        existing_labels = [list(d.keys())[0] for d in in_labels] # this just gets the axes names
        logging.debug('The existing labels are:%s, we will remove:%s' % (existing_labels, PATTERN_LABELS))
        logging.debug('removing these labels from the list')
        core_labels_raw = [l for l in existing_labels if l not in PATTERN_LABELS] # removes them from the list
        core_labels = [l + '.' + in_labels[0][l] for l in core_labels_raw] # add the units in for the ones we are keeping
        # now we just have to add the new ones to this.
        trans_units = self.get_output_axis_units()

        probe_labels = list(core_labels) # take a copy
        probe_labels.extend(['mode_idx.number','x.' + trans_units, 'y.' + trans_units, ])
        logging.debug('the labels for the probe are:%s' % str(probe_labels))
        object_labels = list(core_labels)
        object_labels.extend(['mode_idx.number','x.' + trans_units, 'y.' + trans_units])
        logging.debug('the labels for the object are:%s' % str(object_labels))
        position_labels = list(core_labels)
        position_labels.extend(['xy.m','idx'])
        logging.debug('the labels for the positions are:%s' % str(position_labels))
        # now we also need this part of the shape of the data so...
        md = in_dataset[0].meta_data
        sh = tuple([len(md.get(l)) for l in core_labels_raw])
        return position_labels, probe_labels, object_labels, sh

    def set_probe_rotation_patterns(self, probe, rest_probe):
        try:
            rot_axis = probe.get_data_dimension_by_axis_label('rotation_angle', contains=True) # get the rotation axis
        except Exception as e:
            logging.warning(str(e) + 'we were looking for "rotation_angle"')
            logging.debug('This is not a tomography, so no time series for the probe')
        else:
            # print('the rotation axis is:%s' % str(rot_axis))
            probe_ts = {'core_dims':(rot_axis,),
                        'slice_dims':tuple(set(rest_probe) - set([rot_axis]))}
            probe.add_pattern("TIMESERIES", **probe_ts) # so we can FT the wiggles etc...
            # print('This is a tomography so I have added a TIMESERIES pattern to the probe') # the probe oscillates in time for each projection, set this as a time series pattern

    def set_probe_energy_patterns(self, probe, rest_probe):
        try:
            energy_axis = probe.get_data_dimension_by_axis_label('energy', contains=True) # get an energy axis
        except Exception as e:
            logging.warning(str(e) + 'we were looking for "energy"')
            logging.debug('This is not spectro-microscopy, so no spectrum/timeseries for the probe')
        else:
            probe_spec = {'core_dims':tuple(energy_axis), 'slice_dims':tuple(set(rest_probe) - set([energy_axis]))}
            probe.add_pattern("SPECTRUM", **probe_spec)
            probe.add_pattern("TIMESERIES", **probe_spec)
            logging.debug('This is probably spectro-microscopy so I have added a SPECTRUM pattern to the probe')
            logging.debug('I have also added a TIMESERIES pattern on the same axis, but be careful with what this means!') # the probe oscillates in time for each projection, set this as a time series pattern


    def set_projection_pattern(self, probe, rest_probe):
        probe_proj_core = tuple([rest_probe[idx] for idx in (-3, -2)]) # hard coded since we set them just above
        probe_slice = tuple(set(rest_probe) - set(probe_proj_core))
        probe_proj = {'core_dims':probe_proj_core, 'slice_dims':probe_slice}
        probe.add_pattern("PROJECTION", **probe_proj)
        logging.debug('have added a PROJECTION pattern')

    def set_object_energy_patterns(self, object_trans, rest_obj):
        try:
            energy_axis = object_trans.get_data_dimension_by_axis_label('energy', contains=True) # get an energy axis
        except Exception as e:
            logging.warning(str(e) + 'we were looking for "energy"')
            logging.debug('This is not spectro-microscopy, so no spectrum for the object')
        else:
            obj_spec = {'core_dims':tuple(energy_axis), 'slice_dims':tuple(set(rest_obj) - set([energy_axis]))}
            object_trans.add_pattern("SPECTRUM", **obj_spec)
            logging.debug('This is probably spectro-microscopy so I have added a SPECTRUM pattern to the object') # the probe oscillates in time for each projection, set this as a time series pattern


    def set_object_rotation_patterns(self, object_trans, rest_obj):
        try:
            rot_axis = object_trans.get_data_dimension_by_axis_label('rotation_angle', contains=True) # get the rotation axis
        except Exception as e:
            logging.warning(str(e) + 'we were looking for "rotation_angle"')
            logging.debug('This is not a tomography, so no sinograms for the object transmission')
        else:
            x_axis = object_trans.get_data_dimension_by_axis_label('x', contains=True) # get the x axis
            obj_sino = {'core_dims':(rot_axis, x_axis), 'slice_dims':tuple(set(rest_obj) - set((rot_axis, x_axis)))}
            object_trans.add_pattern("SINOGRAM", **obj_sino) # for the tomography
            logging.debug('This is a tomography so I have added a SINOGRAM pattern to the object transmission') # the probe oscillates in time for each projection, set this as a time series pattern



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: config_utils
   :platform: Unix
   :synopsis: Helper functions for the configurator commands.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
import re
import sys
import os
import mmap
import atexit
import logging
import traceback
import pkgutil

import importlib.util
from functools import wraps
from . import arg_parsers as parsers
import savu.plugins.utils as pu
import savu.data.data_structures.utils as du

if os.name == 'nt':
    from . import win_readline as readline
else:
    import readline

histfile = os.path.join(os.path.expanduser("~"), ".savuhist")
histlen = 1000
logging.basicConfig(level='CRITICAL')
error_level = 0


class DummyFile(object):
    def write(self, x): pass


def _redirect_stdout():
    save_stdout = sys.stdout
    sys.stdout = DummyFile()
    return save_stdout

def load_history_file(hfile):
    try:
        readline.read_history_file(hfile)
        readline.set_history_length(histlen)
    except IOError:
        pass
    atexit.register(write_history_to_file)

def write_history_to_file():
    try:
        readline.write_history_file(histfile)
    except IOError:
        pass

def _set_readline(completer):
    # we want to treat '/' as part of a word, so override the delimiters
    readline.set_completer_delims(' \t\n;')
    readline.parse_and_bind("tab: complete")
    readline.set_completer(completer)


def parse_args(function):
    @wraps(function)
    def _parse_args_wrap_function(content, args):
        doc = function.__doc__
        parser = '%s_arg_parser' % function.__name__
        args = getattr(parsers, parser)(args.split(), doc)
        if not args:
            return content
        return function(content, args)

    return _parse_args_wrap_function


def error_catcher(function):
    @wraps(function)
    def error_catcher_wrap_function(content, args):
        try:
            return function(content, args)
        except Exception as e:
            err_msg_list = str(e).split()
            savu_error = True if len(err_msg_list) > 1 and err_msg_list[1] == 'ERROR:' else False

            if error_level is 0 and savu_error:
                print(e)
            elif error_level is 0:
                print(f"{type(e).__name__}: {e}")
            elif error_level is 1:
                traceback.print_exc(file=sys.stdout)

            return content

    return error_catcher_wrap_function


def populate_plugins(error_mode=False, examples=False):
    # load all the plugins
    plugins_paths = pu.get_plugins_paths(examples=examples)
    failed_imports = {}

    for path, name in plugins_paths.items():
        for finder, module_name, is_pkg in pkgutil.walk_packages([path], name):
            if not is_pkg:
                _load_module(finder, module_name, failed_imports, error_mode)
    

def _load_module(finder, module_name, failed_imports, error_mode):
    try:
        # need to ignore loading of plugin.utils as it is emptying the list
        spec = finder.find_spec(module_name)
        mod = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = mod
        spec.loader.exec_module(mod)
    except Exception as e:
        if _is_registered_plugin(mod):       
            clazz = pu._get_cls_name(module_name)
            failed_imports[clazz] = e
            if error_mode:
                print(("\nUnable to load plugin %s\n%s" % (module_name, e)))

    return failed_imports


def _is_registered_plugin(mod):
    with open(mod.__file__) as f:
        for line in f:
            if "@register_plugin" in line and line.replace(' ', '')[0] != '#':
                return True
    return False


def _dawn_setup():
    for plugin in list(pu.dawn_plugins.keys()):
        p = pu.plugins[plugin]()
        pu.dawn_plugins[plugin]['input rank'] = \
            du.get_pattern_rank(p.get_plugin_pattern())
        pu.dawn_plugins[plugin]['description'] = p.__doc__.split(':param')[0]
        params = _get_dawn_parameters(p)
        pu.dawn_plugin_params[plugin] = params


def _get_dawn_parameters(plugin):
    plugin._populate_default_parameters()
    desc = plugin.parameters_desc
    params = {}
    for key, value in plugin.parameters.items():
        if key not in ['in_datasets', 'out_datasets']:
            params[key] = {'value': value, 'hint': desc[key]}
    return params


def _populate_plugin_list(content, pfilter=""):
    """ Populate the plugin list from a list of plugin instances. """
    content.plugin_list.plugin_list = []
    sorted_plugins = __get_filtered_plugins(pfilter)
    count = 0
    for key in sorted_plugins:
        content.add(key, str(count))
        count += 1


def __get_filtered_plugins(pfilter):
    """ Get a sorted, filter list of plugins. """
    key_list = []
    star_search = \
        pfilter.split('*')[0] if pfilter and '*' in pfilter else False

    for key, value in pu.plugins.items():
        if star_search:
            search = '(?i)^' + star_search
            if re.match(search, value.__name__) or \
                    re.match(search, value.__module__):
                key_list.append(key)
        elif pfilter in value.__module__ or pfilter in value.__name__:
            key_list.append(key)

    key_list.sort()
    return key_list


def __get_start_stop(content, start, stop):
    range_dict = {}
    if start:
        start = content.find_position(start)
        stop = content.find_position(stop) + 1 if stop else start + 1
        range_dict['start'] = start
        range_dict['stop'] = stop
    return range_dict



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_type
   :platform: Unix
   :synopsis: A module to load MRC image data.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
from savu.data.data_structures.data_types.base_type import BaseType
import mrcfile
import numpy as np


class MRC(BaseType):

    def __init__(self, Data, filename, stats=None):
        self._data_obj = Data
        self.filename = filename
        super(MRC, self).__init__()

        self.file = mrcfile.mmap(filename, 'r')
        self.dtype = self.file.data.dtype

    def __getitem__(self, idx):
        data = self.file.data[idx].astype(np.float32)
        data /= 1000.0
        return data

    def get_shape(self):
        return self.file.data.shape

    def clone_data_args(self, args, kwargs, extras):
        args = ['self', 'filename']
        kwargs['stats'] = 'stats'
        return args, kwargs, extras

    def close(self):
        self.file.close()



import argparse
import h5py
import sys
import os

from savu.version import __version__


class NXcitation(object):
    def __init__(self, description, doi, endnote, bibtex):
        self.description = description
        self.doi = doi
        self.endnote = endnote
        self.bibtex = bibtex

    def get_bibtex_ref(self):
        return self.bibtex.split(',')[0].split('{')[1]

    def get_first_author(self):
        parts = self.endnote.split('\n')
        for part in parts:
            if part.startswith("%A"):
                return part.replace("%A", "").strip()

    def get_date(self):
        parts = self.endnote.split('\n')
        for part in parts:
            if part.startswith("%D"):
                return part.replace("%D", "").strip()

    def get_description_with_author(self):
        return "%s \\ref{%s}(%s, %s)" % (self.description,
                                         self.get_bibtex_ref(),
                                         self.get_first_author(),
                                         self.get_date())


class NXcitation_manager(object):
    def __init__(self):
        self.NXcite_list = []

    def add_citation(self, citation):
        self.NXcite_list.append(citation)

    def get_full_endnote(self):
        return "\n\n".join([cite.endnote for cite in self.NXcite_list])

    def get_full_bibtex(self):
        return "\n".join([cite.bibtex for cite in self.NXcite_list])

    def get_description_with_citations(self):
        return ".  ".join([cite.get_description_with_author() for cite in
                           self.NXcite_list])

    def __str__(self):
        return "\nDESCRIPTION\n%s\n\nBIBTEX\n%s\n\nENDNOTE\n%s" % \
            (self.get_description_with_citations(), self.get_full_bibtex(),
             self.get_full_endnote())


class NXciteVisitor(object):

    def __init__(self):
        self.citation_manager = NXcitation_manager()

    def _visit_NXcite(self, name, obj):
        if "NX_class" in list(obj.attrs.keys()):
            if obj.attrs["NX_class"] in ["NXcite"]:
                citation = NXcitation(obj['description'][0],
                                      obj['doi'][0],
                                      obj['endnote'][0],
                                      obj['bibtex'][0])
                self.citation_manager.add_citation(citation)

    def get_citation_manager(self, nx_file, entry):
        nx_file[entry].visititems(self._visit_NXcite)
        return self.citation_manager


def __check_input_params(args):
    """ Check for required input arguments.
    """
    if len(args) is not 2:
        print("Input and output filename need to be specified")
        print("Exiting with error code 1 - incorrect number of inputs")
        sys.exit(1)

    if not os.path.exists(args[0]):
        print(("Input file '%s' does not exist" % args[0]))
        print("Exiting with error code 2 - Input file missing")
        sys.exit(2)


def __option_parser():
    """ Option parser for command line arguments.
    """
    version = "%(prog)s " + __version__
    parser = argparse.ArgumentParser()
    parser.add_argument('in_file', help='Input data file.')
    parser.add_argument('out_file', help='Output file to extract citation \
                        information to.')
    parser.add_argument('--version', action='version', version=version)
    return parser.parse_args()


def main():
    args = __option_parser()
    infile = h5py.File(args.in_file, 'r')
    citation_manager = NXciteVisitor().get_citation_manager(infile, '/')
    if citation_manager is not None:
        with open(args.out_file, 'a') as outfile:
            outfile.write(citation_manager.__str__())
    print("Extraction complete")

if __name__ == '__main__':
    main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: vo_centering_iterative
   :platform: Unix
   :synopsis: A plugin to find the center of rotation per frame
.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>
"""

import math
import logging
import numpy as np
import scipy.ndimage as ndi
import scipy.ndimage.filters as filter
import pyfftw.interfaces.scipy_fftpack as fft

from scipy import signal

from savu.plugins.utils import register_plugin
from savu.data.plugin_list import CitationInformation
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.iterative_plugin import IterativePlugin

#    :u*param search_area: Search area in pixels from horizontal approximate \
#        centre of the image. Default: (-50, 50).
#    Deprecated !!!

@register_plugin
class VoCenteringIterative(BaseFilter, IterativePlugin):
    """
    A plugin to calculate the centre of rotation using the Vo Method

    :param ratio: The ratio between the size of object and FOV of \
        the camera. Default: 0.5.
    :param row_drop: Drop lines around vertical center of the \
        mask. Default: 20.
    :param search_radius: Use for fine searching. Default: 6.
    :param step: Step of fine searching. Default: 0.5.
    :param expand_by: The number of pixels to expand the search region by \
        on each iteration.  Default: 5
    :param boundary_distance: Accepted distance of minima from the boundary of\
        the listshift in the coarse search.  Default: 3.
    :u*param preview: A slice list of required frames (sinograms) to use in \
    the calulation of the centre of rotation (this will not reduce the data \
    size for subsequent plugins). Default: [].
    :param datasets_to_populate: A list of datasets which require this \
        information. Default: [].
    :param out_datasets: The default \
        names. Default: ['cor_raw','cor_fit', 'reliability'].
    :u*param start_pixel: The approximate centre. If value is None, take the \
        value from .nxs file else set to image centre. Default: None.
    """

    def __init__(self):
        super(VoCenteringIterative, self).__init__("VoCenteringIterative")
        self.search_area = (-20, 20)
        self.peak_height_min = 50000  # arbitrary
        self.min_dist = 3  # min distance deamed acceptible from boundary
        self.expand_by = 5  # expand the search region by this amount
        self.list_shift = None
        self.warning_level = 0
        self.final = False
        self.at_boundary = False
        self.list_metric = []
        self.expand_direction = None

    def _create_mask(self, Nrow, Ncol, obj_radius):
        du, dv = 1.0/Ncol, (Nrow-1.0)/(Nrow*2.0*math.pi)
        cen_row, cen_col = int(np.ceil(Nrow / 2.0)-1), int(np.ceil(Ncol / 2.0)-1)
        drop = self.parameters['row_drop']
        mask = np.zeros((Nrow, Ncol), dtype=np.float32)
        for i in range(Nrow):
            num1 = np.round(((i-cen_row)*dv/obj_radius)/du)
            p1, p2 = (np.clip(np.sort((-num1+cen_col, num1+cen_col)),
                              0, Ncol-1)).astype(int)
            mask[i, p1:p2+1] = np.ones(p2-p1+1, dtype=np.float32)

        if drop < cen_row:
            mask[cen_row-drop:cen_row+drop+1, :] = \
                np.zeros((2*drop + 1, Ncol), dtype=np.float32)
        mask[:, cen_col-1:cen_col+2] = np.zeros((Nrow, 3), dtype=np.float32)
        return mask

    def _get_start_shift(self, centre):
        if self.parameters['start_pixel'] is not None:
            shift = centre - int(self.parameters['start_pixel']/self.downlevel)
        else:
            in_mData = self.get_in_meta_data()[0]
            shift = centre - in_mData['centre'] if 'centre' in \
                list(in_mData.get_dictionary().keys()) else 0
        return int(shift)

    def _coarse_search(self, sino, list_shift):
        # search minsearch to maxsearch in 1 pixel steps
        list_metric = np.zeros(len(list_shift), dtype=np.float32)
        (Nrow, Ncol) = sino.shape
        # check angles to determine if a sinogram should be chopped off.
        # Copy the sinogram and flip left right, to make a full [0:2Pi] sino
        sino2 = np.fliplr(sino[1:])
        # This image is used for compensating the shift of sino2
        compensateimage = np.zeros((Nrow-1, Ncol), dtype=np.float32)
        # Start coarse search in which the shift step is 1
        compensateimage[:] = np.flipud(sino)[1:]
        mask = self._create_mask(2*Nrow-1, Ncol,
                                 0.5*self.parameters['ratio']*Ncol)
        count = 0
        for i in list_shift:
            sino2a = np.roll(sino2, i, axis=1)
            if i >= 0:
                sino2a[:, 0:i] = compensateimage[:, 0:i]
            else:
                sino2a[:, i:] = compensateimage[:, i:]
            list_metric[count] = np.sum(
                np.abs(fft.fftshift(fft.fft2(np.vstack((sino, sino2a)))))*mask)
            count += 1
        return list_metric

    def _fine_search(self, sino, raw_cor):
        (Nrow, Ncol) = sino.shape
        centerfliplr = (Ncol + 1.0) / 2.0 - 1.0
        # Use to shift the sino2 to the raw CoR
        shiftsino = np.int16(2*(raw_cor-centerfliplr))
        sino2 = np.roll(np.fliplr(sino[1:]), shiftsino, axis=1)
        lefttake = 0
        righttake = Ncol-1
        search_rad = self.parameters['search_radius']

        if raw_cor <= centerfliplr:
            lefttake = np.int16(np.ceil(search_rad+1))
            righttake = np.int16(np.floor(2*raw_cor-search_rad-1))
        else:
            lefttake = np.int16(np.ceil(raw_cor-(Ncol-1-raw_cor)+search_rad+1))
            righttake = np.int16(np.floor(Ncol-1-search_rad-1))

        Ncol1 = righttake-lefttake + 1
        mask = self._create_mask(2*Nrow-1, Ncol1,
                                 0.5*self.parameters['ratio']*Ncol)
        numshift = np.int16((2*search_rad)/self.parameters['step'])+1
        listshift = np.linspace(-search_rad, search_rad, num=numshift)
        listmetric = np.zeros(len(listshift), dtype=np.float32)
        num1 = 0
        factor1 = np.mean(sino[-1, lefttake:righttake])
        for i in listshift:
            sino2a = ndi.interpolation.shift(sino2, (0, i), prefilter=False)
            factor2 = np.mean(sino2a[0, lefttake:righttake])
            sino2a = sino2a*factor1/factor2
            sinojoin = np.vstack((sino, sino2a))
            listmetric[num1] = np.sum(np.abs(fft.fftshift(
                fft.fft2(sinojoin[:, lefttake:righttake + 1])))*mask)
            num1 = num1 + 1
        minpos = np.argmin(listmetric)
        rotcenter = raw_cor + listshift[minpos] / 2.0
        return rotcenter

    def _get_listshift(self):
        smin, smax = self.search_area if self.get_iteration() is 0 \
            else self._expand_search()
        list_shift = np.arange(smin, smax+2, 2) - self.start_shift
        logging.debug('list shift is %s', list_shift)
        return list_shift

    def _expand_search(self):
        if self.expand_direction == 'left':
            return self._expand_left()
        elif self.expand_direction == 'right':
            return self._expand_right()
        else:
            raise Exception('Unknown expand direction.')

    def _expand_left(self):
        smax = self.list_shift[0] - 2
        smin = smax - self.expand_by*2

        if smin <= -self.boundary:
            smin = -self.boundary
            self.at_boundary = True
        return smin, smax

    def _expand_right(self):
        smin = self.list_shift[-1] + 2
        smax = self.list_shift[-1] + self.expand_by*2

        if smax <= self.boundary:
            smax = self.boundary
            self.at_boundary = True

        return smin, smax

    def pre_process(self):
        pData = self.get_plugin_in_datasets()[0]
        label = pData.get_data_dimension_by_axis_label
        Ncol = pData.get_shape()[label('detector_x')]
        self.downlevel = 4 if Ncol > 1800 else 1
        self.downsample = slice(0, Ncol, self.downlevel)
        Ncol_downsample = len(np.arange(0, Ncol, self.downlevel))
        self.centre_fliplr = (Ncol_downsample - 1.0) / 2.0
        self.start_shift = self._get_start_shift(self.centre_fliplr)*2
        self.boundary = int(np.ceil(Ncol/4.0))

    def process_frames(self, data):
        if not self.final:
            logging.debug('performing coarse search for iteration %s',
                          self.get_iteration())
            sino = filter.gaussian_filter(data[0][:, self.downsample], (3, 1))
            list_shift = self._get_listshift()
            list_metric = self._coarse_search(sino, list_shift)
            self._update_lists(list(list_shift), list(list_metric))

            self.coarse_cor, dist, reliability_metrics = \
                self._analyse_result(self.list_metric, self.list_shift)

            return [np.array([self.coarse_cor]), np.array([dist]),
                    np.array([reliability_metrics]), np.array([self.list_metric])]
        else:
            logging.debug("performing fine search")
            sino = filter.median_filter(data[0], (2, 2))
            cor = self._fine_search(sino, self.coarse_cor)
            self.set_processing_complete()
            return [np.array([cor]), np.array([self.list_metric])]

    def _update_lists(self, shift, metric):
        if self.expand_direction == 'left':
            self.list_shift = shift + self.list_shift
            self.list_metric = metric + self.list_metric
        elif self.expand_direction == 'right':
            self.list_shift += shift
            self.list_metric += metric
        else:
            self.list_shift = shift
            self.list_metric = metric

    def _analyse_result(self, metric, shift):
        minpos = np.argmin(metric)
        dist = min(abs(len(shift) - minpos), -minpos)

        rot_centre = (self.centre_fliplr + shift[minpos] / 2.0)*self.downlevel
        peaks = self._find_peaks(metric)

        good_nPeaks = True
        if len(peaks) != 1:
            good_nPeaks = False
        good_peak_height = True if np.any(peaks) and \
            max(peaks) > self.peak_height_min else False

        metric = 0.0
        if (good_peak_height and good_nPeaks):
            metric = 1.0
        elif (good_peak_height or good_nPeaks):
            metric = 0.5

        return rot_centre, dist, metric

    def _find_peaks(self, metric):
        import peakutils
        grad2 = np.gradient(np.gradient(metric))
        grad2[grad2 < 0] = 0
        index = peakutils.indexes(grad2, thres=0.5, min_dist=3)
        return np.sort(grad2[index])

    def post_process(self):
        logging.debug("in the post process function")
        in_datasets, out_datasets = self.get_datasets()

        # =====================================================================
        # Analyse distance of centre values from boundary of search region
        dist_from_boundary = np.squeeze(out_datasets[1].data[...])
        near_boundary = np.where(abs(dist_from_boundary) < self.min_dist)[0]
        nEntries = len(dist_from_boundary)

        # Case1: Greater than half the results are near the boundary
        if (len(near_boundary)/float(nEntries)) > 0.5:
            # find which boundary
            signs = np.sign(dist_from_boundary[near_boundary])
            left, right = len(signs[signs < 0]), len(signs[signs > 0])

            logging.debug("res: results are near boundary")
            if not self.at_boundary:
                # if they are all at the same boundary expand the search region
                if not (left and right):
                    logging.debug("res: expanding")
                    self.expand_direction = 'left' if left else 'right'
                # if they are at different boundaries determine which values
                # are most reliable
                else:
                    logging.debug("res: choosing a boundary")
                    self.expand_direction = \
                        self._choose_boundary(near_boundary, signs)
                    # case that the results are close to different boundaries
                    # Analyse reliability and choose direction
            else:
                logging.debug("res: at the edge of the boundary")
                # Move on to the fine search
                self._set_final_process()
                self.warning_level = 1 # change this to be more descriptive ***
        else:
            logging.debug("result is not near the boundary")
            # Move on to the fine search
            self._set_final_process()
        # =====================================================================

    def _choose_boundary(self, idx, signs):
        good, maybe, bad = self._get_reliability_levels()
        sign = self._check_entries(good, signs[good])
        self.warning_level = 0
        if not sign:
            sign = self._check_entries(maybe, signs[maybe])
            self.warning_level = 1
        if not sign:
            sign = self._check_entries(bad, signs[bad])
            self.warning_level = 2
        return sign

    def _check_entries(self, idx, signs):
        if np.any(idx):
            left, right = signs[signs < 0], signs[signs > 0]
            if not (left and right):
                # use all the good ones
                return 'left' if left else 'right'
        return None

    def _get_reliability_levels(self, final=False):
        in_datasets, out_datasets = \
            self.get_datasets() if not final else self.get_original_datasets()
        reliability = np.squeeze(out_datasets[2].data[...])
        logging.debug('reliability is %s', reliability)
        good = np.where(reliability == 1.0)[0]
        maybe = np.where(reliability == 0.5)[0]
        bad = np.where(reliability == 0.0)[0]
        return good, maybe, bad

    def final_post_process(self):

        # choose which values to include
        good, maybe, bad = self._get_reliability_levels(final=True)
        # Do I need to change the warning levels here?
        entries = good if np.any(good) else maybe if np.any(maybe) else bad
        self.warning_level = 0 if np.any(good) else 1 if np.any(maybe) else 2
        logging.debug('sinograms used in final calculations are %s', entries)

        # do some curve fitting here
        # Get a handle on the original datasets
        in_dataset, out_dataset = self.get_original_datasets()
        cor_raw = np.squeeze(out_dataset[0].data[...])[entries]
        cor_fit = out_dataset[1].data[...]
        fit = np.zeros(cor_fit.shape)
        fit[:] = np.median(cor_raw)
        cor_fit = fit
        out_dataset[1].data[:] = cor_fit[:]

        self.populate_meta_data('cor_raw', cor_raw)
        self.populate_meta_data('centre_of_rotation',
                                out_dataset[1].data[:].squeeze(axis=1))

    def _set_final_process(self):
        self.final = True
        self.post_process = self.final_post_process
        in_dataset, out_dataset = self.get_datasets()
        self.set_iteration_datasets(
                self.get_iteration()+1, [in_dataset[0]], [out_dataset[0]])

    def populate_meta_data(self, key, value):
        datasets = self.parameters['datasets_to_populate']
        in_meta_data = self.get_in_meta_data()[0]
        in_meta_data.set(key, value)
        for name in datasets:
            self.exp.index['in_data'][name].meta_data.set(key, value)

    def setup(self):
        # set up the output dataset that is created by the plugin
        in_dataset, out_dataset = self.get_datasets()

        self.orig_full_shape = in_dataset[0].get_shape()

        # reduce the data as per data_subset parameter
        self.set_preview(in_dataset[0], self.parameters['preview'])

        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', self.get_max_frames())
        # copy all required information from in_dataset[0]
        fullData = in_dataset[0]

        slice_dirs = np.array(in_dataset[0].get_slice_dimensions())
        new_shape = (np.prod(np.array(fullData.get_shape())[slice_dirs]), 1)
        self.orig_shape = \
            (np.prod(np.array(self.orig_full_shape)[slice_dirs]), 1)

        self._create_metadata_dataset(out_dataset[0], new_shape)
        self._create_metadata_dataset(out_dataset[1], self.orig_shape)
        self._create_metadata_dataset(out_dataset[2], new_shape)

        # output metric
        new_shape = (np.prod(np.array(fullData.get_shape())[slice_dirs]), 21)
        self._create_metadata_dataset(out_dataset[3], new_shape)

        out_pData[0].plugin_data_setup('METADATA', self.get_max_frames())
        out_pData[1].plugin_data_setup('METADATA', self.get_max_frames())
        out_pData[2].plugin_data_setup('METADATA', self.get_max_frames())
        out_pData[3].plugin_data_setup('METADATA', self.get_max_frames())

    def _create_metadata_dataset(self, data, shape):
        data.create_dataset(shape=shape,
                            axis_labels=['x.pixels', 'y.pixels'],
                            remove=True,
                            transport='hdf5')
        data.add_pattern("METADATA", core_dims=(1,), slice_dims=(0,))

    def nOutput_datasets(self):
        return 4

    def get_max_frames(self):
        return 'single'

    def fix_transport(self):
        # This plugin requires communication between processes in the post
        # process, which it does via files
        return 'hdf5'

    def executive_summary(self):
        if self.warning_level == 0:
            msg = "Confidence in the centre value is high."
        elif self.warning_level == 1:
            msg = "Confidence in the centre value is average."
        else:
            msg = "Confidence in the centre value is low."
        return [msg]

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The center of rotation for this reconstruction was calculated " +
             "automatically using the method described in this work")
        cite_info.bibtex = \
            ("@article{vo2014reliable,\n" +
             "title={Reliable method for calculating the center of rotation " +
             "in parallel-beam tomography},\n" +
             "author={Vo, Nghia T and Drakopoulos, Michael and Atwood, " +
             "Robert C and Reinhard, Christina},\n" +
             "journal={Optics Express},\n" +
             "volume={22},\n" +
             "number={16},\n" +
             "pages={19078--19086},\n" +
             "year={2014},\n" +
             "publisher={Optical Society of America}\n" +
             "}")
        cite_info.endnote = \
            ("%0 Journal Article\n" +
             "%T Reliable method for calculating the center of rotation in " +
             "parallel-beam tomography\n" +
             "%A Vo, Nghia T\n" +
             "%A Drakopoulos, Michael\n" +
             "%A Atwood, Robert C\n" +
             "%A Reinhard, Christina\n" +
             "%J Optics Express\n" +
             "%V 22\n" +
             "%N 16\n" +
             "%P 19078-19086\n" +
             "%@ 1094-4087\n" +
             "%D 2014\n" +
             "%I Optical Society of America")
        cite_info.doi = "https://doi.org/10.1364/OE.22.019078"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: simple_fit
   :platform: Unix
   :synopsis: A plugin to fit peaks

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""
import logging
from savu.plugins.utils import register_plugin
from savu.plugins.fitters.base_fitter import BaseFitter
import numpy as np
from scipy.optimize import leastsq
import time
import math

@register_plugin
class SimpleFit(BaseFitter):
    """
    This plugin fits peaks.
    :param width_guess: An initial guess at the width. Default: 0.02.
    :param PeakIndex: the peak index. Default: [].
    """

    def __init__(self):
        super(SimpleFit, self).__init__("SimpleFit")

    def pre_process(self):
        in_meta_data = self.get_in_meta_data()[0]
        out_meta_data = self.get_out_meta_data()[0]
        in_meta_data.set("PeakIndex", self.parameters["PeakIndex"])        
        self.axis = in_meta_data.get("Q")
        self.peakindex = in_meta_data.get("PeakIndex")
        self.positions = self.axis[self.peakindex]
        in_meta_data.set('PeakQ', self.positions)

    def process_frames(self, data):
        t1 = time.time()
        data = data[0]
        axis = self.axis
        positions = self.positions
        #print positions
        weights = data[self.peakindex]
        widths = np.ones_like(positions)*self.parameters["width_guess"]
        p = []
        p.extend(weights)
        p.extend(widths)
        curvetype = self.getFitFunction(str(self.parameters['peak_shape']))
        [x2, cov_x2, infodict2, msg2, ier2] = leastsq(self._resid, p,
                       args=(curvetype, data, axis, positions),
                       Dfun=self.dfunc, col_deriv=1, full_output=1)
        logging.debug("done one")
        params = x2
        if np.isnan(params).any():
            logging.debug('Nans were detected here')
            params = np.zeros(len(params))
        #print msg2
        weights, widths, areas = self.getAreas(curvetype,
                                               axis, positions, params)
        residuals = self._resid(params, curvetype, data, axis, positions)
        # all fitting routines will output the same format.
        # nchannels long, with 3 elements. Each can be a subarray.
        t2 = time.time()
        logging.debug("Simple fit iteration took: %s ms", str((t2-t1)*1e3))
        return [weights, widths, areas, residuals]

    def setup(self):
        # set up the output datasets that are created by the plugin
        in_dataset, out_datasets = self.get_datasets()

        shape = in_dataset[0].get_shape()
        axis_labels = ['-1.PeakIndex.pixel.unit']
        pattern_list = ['SINOGRAM', 'PROJECTION']

        fitAreas = out_datasets[0]
        fitHeights = out_datasets[1]
        fitWidths = out_datasets[2]
        peakindex = self.parameters['PeakIndex']
        new_shape = shape[:-1] + (len(peakindex),)

        fitAreas.create_dataset(patterns={in_dataset[0]: pattern_list},
                                axis_labels={in_dataset[0]: axis_labels},
                                shape=new_shape)

        fitHeights.create_dataset(patterns={in_dataset[0]: pattern_list},
                                  axis_labels={in_dataset[0]: axis_labels},
                                  shape=new_shape)

        fitWidths.create_dataset(patterns={in_dataset[0]: pattern_list},
                                 axis_labels={in_dataset[0]: axis_labels},
                                 shape=new_shape)

        channel = {'core_dims': (-1,), 'slice_dims': list(range(len(shape)-1))}

        fitAreas.add_pattern("CHANNEL", **channel)
        fitHeights.add_pattern("CHANNEL", **channel)
        fitWidths.add_pattern("CHANNEL", **channel)

        residuals = out_datasets[3]
        residuals.create_dataset(in_dataset[0])

        # setup plugin datasets
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SPECTRUM', self.get_max_frames())

        out_pData[0].plugin_data_setup('CHANNEL', self.get_max_frames())
        out_pData[1].plugin_data_setup('CHANNEL', self.get_max_frames())
        out_pData[2].plugin_data_setup('CHANNEL', self.get_max_frames())
        out_pData[3].plugin_data_setup('SPECTRUM', self.get_max_frames())



from . import GraphicalThreadProfiler as GTP
from . import GraphicalThreadProfiler_multi as GTP_m
import fnmatch
import os


def convert(files):
    import pandas as pd
    # calculate the mean and std

    test = []
    index = ['file_system', 'nNodes_x_nCores', 'Mean_time', 'nNodes',
             'Std_time']
    for file in files:
        temp_frame = pd.read_csv(file, header=None)
        a = (file.split('/')[-1]).split('_')

        vals = pd.Series([a[3],
                         int(a[-6].split('N')[1])*int(a[-5].split('C')[1]),
                         int(temp_frame.iloc[1, 1])*0.001,
                         int(a[-6].split('N')[1]),
                         int(temp_frame.iloc[2, 1])*0.001], index=index)
        test.append(vals)

    all_vals = (pd.concat(test, axis=1).transpose())

    frame = all_vals

    return frame


def get_files(dir_path):
    from os import listdir
    from os.path import isfile, join

    all_files = [f for f in listdir(dir_path) if isfile(join(dir_path, f))]
    files = [(dir_path + '/' + f) for f in all_files]

    return files


def render_template(frame, outfilename, title, size, params, header_shift,
                    max_std):
    from jinja2 import Template
    from . import template_strings as ts

    nVals = len(frame)

    f_out = open(outfilename, 'w')
    template = Template(ts.set_template_string_vis(nVals, title, size, params,
                                                   header_shift))

    style = os.path.dirname(__file__) + '/style_sheet.css'
    print(outfilename)
    f_out.write(template.render(frame=[list(map(list, f)) for f in frame],
                                style_sheet=style, max_bubble=max_std))
    f_out.close()

    return


def convert_all_files():
    all_files = get_files(os.getcwd())
    single_files = [f for f in all_files if f.split('.')[-1][0] is 'o']
    GTP.convert(single_files)

    wildcard_files = [(os.path.dirname(f) + '/' +
                       os.path.basename(f).split('.')[-2] + '*')
                      for f in single_files]
    for wildcard in set(wildcard_files):
        matching_files = []
        for file in single_files:
            if fnmatch.fnmatch(file, wildcard):
                matching_files.append(file)
        GTP_m.convert(matching_files)

    create_bubble_chart(get_files(os.getcwd()))


def create_bubble_chart(all_files):

    stats_files = [f for f in all_files if 'stats.csv' in f]
    frame = convert(stats_files)

    max_std = frame.Std_time.max()

    frame['link'] = [('file://' + f.split('_stats')[0] + '.html')
                     for f in stats_files]

    size = [(70, 70), (100, 100)]
    #params = {}
    params = {'Chunk': 'false', 'Process': '12', 'Data size': '(91,135,160)'}
    render_template([frame.values.tolist()], 'analysis.html', ['Nodes'], size,
                    params, 0, max_std)


if __name__ == "__main__":
    import optparse
    usage = "%prog [options] input_file"

    parser = optparse.OptionParser(usage=usage)

    (options, args) = parser.parse_args()

    if len(args) is 1:
        filename = (os.getcwd() if args[0] is '.' else args[0])
        create_bubble_chart(get_files(filename))
    else:
        convert_all_files()




# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Calculate geodesic distance transform for 3D volume
   :platform: Unix
   :synopsis: Wraps the code for calculating geodesic distance transforms, can be a \
   usefull tool for data segmentation with a proper seed initialisation

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.multi_threaded_plugin import MultiThreadedPlugin
from savu.plugins.utils import register_plugin

"""
Geodesic transformation of images can be implementated with two approaches: \
fast marching and raster scan. Fast marching is based on the iterative \
propagation of a pixel front with velocity F. Raster scan is based on  \
kernel operations that are sequentially applied over the image in multiple passes.\
In GeoS, the authors proposed to use a 3x3 kernel for forward and \
backward passes for efficient geodesic distance transform, which was used for  \
image segmentation. # https://github.com/taigw/geodesic_distance \
In order to make code work one need to specify valid coordinates to initialise the point\
(seed) from which the distances will be calculated.

lambda: weighting betwween 0.0 and 1.0
          if lambda==0.0, return spatial euclidean distance without considering gradient
          if lambda==1.0, the distance is based on gradient only without using spatial distance
"""
import GeodisTK
import numpy as np

@register_plugin
class GeoDistance3d(Plugin, MultiThreadedPlugin):
    """
    3D geodesic transformation of volumes with mask initialisation.

    :param lambda: weighting betwween 0 and 1 . Default: 0.5.
    :param iterations: number of iteration for raster scanning . Default: 4.
    :param out_datasets: The default names . Default: ['GeoDist'].
    """

    def __init__(self):
        super(GeoDistance3d, self).__init__("GeoDistance3d")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        
        # If VOLUME_3D pattern doesn't exist then use "VOlUME_XZ" pattern with
        # all of voxel_y dimension as this is equivalent to one VOLUME_3D scan.
        getall = ['VOLUME_XZ', 'voxel_y']
        in_pData[0].plugin_data_setup('VOLUME_3D', 'single', getall=getall)
        in_pData[1].plugin_data_setup('VOLUME_3D', 'single', getall=getall) # the mask initialisation

        out_dataset[0].create_dataset(in_dataset[0])
        out_pData[0].plugin_data_setup('VOLUME_3D', 'single', getall=getall)

    def pre_process(self):
        # extract given parameters
        self.lambda_par = self.parameters['lambda']
        self.iterations = self.parameters['iterations']

    def process_frames(self, data):
        input_temp = data[0]
        indices = np.where(np.isnan(input_temp))
        input_temp[indices] = 0.0
        spacing = [1.0, 1.0, 1.0]
        if (np.sum(data[1]) > 0):
            geoDist = GeodisTK.geodesic3d_raster_scan(input_temp, data[1], spacing, self.lambda_par, self.iterations)
        else:
            geoDist = np.float32(np.zeros(np.shape(data[0])))
        return geoDist

    def nInput_datasets(self):
        return 2
    def nOutput_datasets(self):
        return 1



import os


def get_all_files_from(folder):
    all_files = []
    for root, dirs, files in os.walk(folder, topdown=True):
        files = [os.path.join(root[root.index(folder)+1+len(folder):], file) for file in files]
        all_files.extend(files)
    return all_files


def get_test_process_list(folder):
    test_process_list = []
    for root, dirs, files in os.walk(folder, topdown=True):
        files[:] = [fi for fi in files if fi.split('.')[-1] == 'nxs']
        # since there are some nxs files inside the subfolders we attach the subfolder
        # name to nxs without the root folder
        files = [os.path.join(root[root.index(folder)+1+len(folder):], file) for file in files]
        test_process_list.extend(files)
    return test_process_list


def get_process_list(folder, search=False):
    process_list = []
    plugin_list = []
    exclude_dir = ['__pycache__']
    exclude_file = ['__init__.py']
    for root, dirs, files in os.walk(folder, topdown=True):
        dirs[:] = [d for d in dirs if d not in exclude_dir]
        files[:] = [fi for fi in files if fi.split('.')[-1] == 'py']
        files[:] = [fi for fi in files if fi not in exclude_file]
        processes = get_process_list_in_file(root, files)
        plugins = get_no_process_list_tests(root, files)
        for p in processes:
            process_list.append(p)
        for p in plugins:
            plugin_list.append(p)
    return process_list, plugin_list


def get_process_list_in_file(root, files):
    processes = []
    for fname in files:
        fname = root + '/' + fname
        with open(fname, 'r') as in_file:
            for line in in_file:
                if '.nxs' in line:
                    processes.append(get_nxs_file_name(line))
    return processes

def get_no_process_list_tests(root, files):
    processes = []
    for fname in files:
        fname = root + '/' + fname
        with open(fname, 'r') as in_file:
            func = 'run_protected_plugin_runner_no_process_list'
            exclude = ['def', 'search_str']
            pos = 1
            param = get_param_name(func, pos, in_file, exclude=exclude)
            if param:
                in_file.seek(0)
                plugin_id_list = get_param_value_from_file(param, in_file)
                for pid in plugin_id_list:
                    plugin_name = pid.split('.')[-1].split("'")[0]
                    processes.append(plugin_name + '.py')
    return processes


def get_nxs_file_name(line):
    split_list = line.split("'")
    for entry in split_list:
        if 'nxs' in entry:
            if (len(entry.split(' ')) == 1):
                return entry


def get_param_name(func, pos, in_file, exclude=[]):
    """ Find the name of an argument passed to a function.

    :param str func: function name
    :param int pos: function argument position
    :param file in_file: open file to search
    :keyword list[str] exclude: ignore lines containing these strings.
                                Defaults to [].
    :returns: name associated with function argument
    :rtype: str
    """
    search_str = 'run_protected_plugin_runner_no_process_list('
    ignore = ['def', 'search_str']
    val_str = None
    for line in in_file:
        if search_str in line:
            if not [i in line for i in ignore].count(True):
                if ')' not in line:
                    line += next(in_file)
                params = line.split('(')[1].split(')')[0]
                val_str = params.split(',')[1].strip()
    return val_str


def get_param_value_from_file(param, in_file):
    """ Find all values associated with a parameter name in a file.

    :param str param: parameter name to search for
    :param file in_file: open file to search
    :returns: value associated with param
    :rtype: list[str]
    """
    param_list = []
    for line in in_file:
        if param in line and line.split('=')[0].strip() == param:
            if "\\" in line:
                line += next(in_file)
                line = ''.join(line.split('\\'))
            value = line.split('=')[1].strip()
            param_list.append(value)
    return param_list



# Copyright 2015 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
.. module:: utils
   :platform: Unix
   :synopsis: Simple core utility methods.
.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

from __future__ import print_function, division

import itertools
import logging
import logging.handlers as handlers

from mpi4py import MPI
from typing import Union


def ensure_string(string: Union[str, bytes]) -> str:
    """
    Python 3 and HDF5 mixture has created a somewhat ambiguous situation
    where the strings in Python 3 are UTF-8 by default,
    but HDF5 strings cannot be unicode.

    This function should wrap strings coming in from HDF5 to ensure that they
    are decoded to UTF-8 so that they can be treated at `str` in Python, and
    not `bytes`.
    """
    if isinstance(string, bytes):
        return string.decode("ascii")
    return string


def logfunction(func):
    """ Decorator to add logging information around calls for use with . """

    def _wrapper(*args, **kwds):
        logging.info("Start::%s:%s",
                     func.__module__,
                     func.__name__)
        returnval = func(*args, **kwds)
        logging.info("Finish::%s:%s",
                     func.__module__,
                     func.__name__)
        return returnval

    return _wrapper


def logmethod(func):
    """ Decorator to add logging information around calls for use with . """

    def _wrapper(self, *args, **kwds):
        logging.info("Start::%s.%s:%s",
                     func.__module__,
                     self.__class__.__name__,
                     func.__name__)
        returnval = func(self, *args, **kwds)
        logging.info("Finish::%s.%s:%s",
                     func.__module__,
                     self.__class__.__name__,
                     func.__name__)
        return returnval

    return _wrapper


def docstring_parameter(*sub):
    """ Decorator to add strings to a doc string."""

    def dec(obj):
        obj.__doc__ = obj.__doc__.format(*sub)
        return obj

    return dec


def import_class(class_name):
    """ Import a class.

    :params: class name
    :returns: class instance
    :rtype: instance of class_name
    """
    name = class_name
    mod = __import__(name)
    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    temp = name.split('.')[-1]
    module2class = ''.join(x.capitalize() for x in temp.split('_'))
    return getattr(mod, module2class.split('.')[-1])


def add_base(this, base):
    """ Add a base class to a class.

    :params class this: a class instance
    :params class base: a class to add as a base class
    """
    cls = this.__class__
    namespace = this.__class__.__dict__.copy()
    this.__class__ = cls.__class__(cls.__name__, (cls, base), namespace)
    base().__init__()


def add_base_classes(this, bases):
    """ Add multiple base classes to a class.

    :params class this: a class instance.
    :params list(class) bases: a list of base classes
    """
    bases = bases if isinstance(bases, list) else [bases]
    for base in bases:
        add_base(this, base)


def get_available_gpus():
    try:
        import pynvml as pv
    except:
        logging.debug("pyNVML module not found")
        raise Exception("pyNVML module not found")
    pv.nvmlInit()
    count = int(pv.nvmlDeviceGetCount())
    return pv, count


USER_LOG_LEVEL = 100
USER_LOG_HANDLER = None


def user_message(message):
    logging.log(USER_LOG_LEVEL, message)
    if USER_LOG_HANDLER is not None:
        USER_LOG_HANDLER.flush()


def user_messages_from_all(header, message_list):
    comm = MPI.COMM_WORLD
    messages = comm.gather(message_list, root=0)
    if messages is None:
        return
    # flatten the list
    messages = list(itertools.chain(*messages))
    if comm.rank == 0:
        for message in set(messages):
            user_message("%s : %i processes report : %s" %
                         (header, messages.count(message), message))


def _output_summary(mpi_flag, plugin):
    if mpi_flag:
        user_messages_from_all(plugin.name, plugin.executive_summary())
    else:
        for message in plugin.executive_summary():
            user_message("%s - %s" % (plugin.name, message))


def add_user_log_level():
    logging.addLevelName(USER_LOG_LEVEL, "USER")


def add_user_log_handler(logger, user_log_path):
    fh = logging.FileHandler(user_log_path, mode='w')
    fh.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    fh.setLevel(USER_LOG_LEVEL)
    logger.addHandler(fh)
    USER_LOG_HANDLER = fh
    user_message("User Log Started")
    user_message("User Log location is '%s'" % (user_log_path))


def add_syslog_log_handler(logger, syslog_address, syslog_port):
    syslog = handlers.SysLogHandler(address=(syslog_address, syslog_port))
    syslog.setFormatter(logging.Formatter('SAVU:%(message)s'))
    syslog.setLevel(logging.WARN)  # only log user log messages
    logger.addHandler(syslog)


def _get_log_level(options):
    """ Gets the right log level for the flags -v or -q
    """
    if ('verbose' in options) and options['verbose']:
        return logging.DEBUG
    if ('quiet' in options) and options['quiet']:
        return logging.WARN
    return logging.INFO


def _send_email(address):
    import smtplib
    from email.mime.text import MIMEText

    me = 'nicola.wadeson@diamond.ac.uk'
    you = address
    # Open a plain text file for reading.  For this example, assume that
    # the text file contains only ASCII characters.
    #    fp = open(textfile, 'rb')
    #    # Create a text/plain message
    #    msg = MIMEText(fp.read())
    #    fp.close()

    # me == the sender's email address
    # you == the recipient's email address
    msg['Subject'] = 'Your Savu job has completed'
    msg['From'] = me
    msg['To'] = you

    # Send the message via our own SMTP server, but don't include the
    # envelope header.
    s = smtplib.SMTP('localhost')
    s.sendmail(me, [you], msg.as_string('write something here'))
    s.quit()


def _savu_encoder(data):
    return f'#savu_encoded#{data}'.encode("ascii")

def _savu_decoder(data):
    data = ensure_string(data)
    if isinstance(data, str) and len(data.split('#savu_encoded#')) > 1:
        return eval(data.split('#savu_encoded#')[-1])

def get_memory_usage_linux(kb=False, mb=True):
    """
    :param kb: Return the value in Kilobytes
    :param mb: Return the value in Megabytes
    :return: The string of the value in either KB or MB
    :rtype str
    """

    try:
        # Windows doesn't seem to have resource package, so this will
        # silently fail
        import resource as res
    except ImportError:
        return 0, 0

    if kb:
        return int(res.getrusage(res.RUSAGE_SELF).ru_maxrss)

    if mb:
        return int(res.getrusage(res.RUSAGE_SELF).ru_maxrss) // 1024



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: nxstxm_loader
   :platform: Unix
   :synopsis: A class for loading nxstxm data

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.loaders.mapping_loaders.base_multi_modal_loader \
    import BaseMultiModalLoader
import numpy as np
from savu.plugins.utils import register_plugin
import logging


@register_plugin
class NxptychoLoader(BaseMultiModalLoader):
    """
    A class to load tomography data from an NXstxm file
    :param name: The name assigned to the dataset. Default: 'ptycho'.
    """

    def __init__(self, name='NxptychoLoader'):
        super(NxptychoLoader, self).__init__(name)

    def setup(self):
        """
         Define the input nexus file

        :param path: The full path of the NeXus file to load.
        :type path: str
        """

        data_str = '/instrument/detector/data'
        data_obj, stxm_entry = self.multi_modal_setup('NXptycho', data_str,
                                                      self.parameters['name'], patterns=False)
        mono_energy = data_obj.backing_file[
            stxm_entry.name + '/instrument/monochromator/energy']
        self.exp.meta_data.set("mono_energy", mono_energy)

        labels = []
        ### set the rotation
        rotation_angle = \
            data_obj.backing_file[stxm_entry.name + '/data/theta'][()]
        if rotation_angle.ndim > 1:
            rotation_angle = rotation_angle[:, 0]
        # axis label
        labels.append('rotation_angle.degrees')

        data_obj.meta_data.set('rotation_angle', rotation_angle)

        ### set the x
        x = \
            data_obj.backing_file[stxm_entry.name + '/data/lab_sxy/lab_sx'][()]*1e-6
        data_obj.meta_data.set('x', x)
        # axis label

        ### set the y
        y = \
            data_obj.backing_file[stxm_entry.name + '/data/lab_sxy/lab_sy'][()]*1e-6
        pos = np.zeros((2,len(y)))
        pos[0,:] = y
        pos[1,:] = x
        data_obj.meta_data.set('xy', pos)
        # axis label
        labels.append('xy.metres')
#         labels.append('y.microns')
        # now set them
        labels.extend(['detectorX.pixel','detectorY.pixel'])
        #print labels
        data_obj.set_axis_labels(*tuple(labels))


        dims = list(range(len(data_obj.get_shape())))
        diff_core = (-2,-1) # it will always be this
        #print diff_core

        diff_slice = tuple(dims[:-2])
        #print diff_slice
        logging.debug("is a diffraction")
        logging.debug("the diffraction cores are:"+str(diff_core))
        logging.debug("the diffraction slices are:"+str(diff_slice))
        data_obj.add_pattern("DIFFRACTION", core_dims=diff_core,
                             slice_dims=diff_slice)

        data_obj.add_pattern("4D_SCAN", core_dims=(1,2,3),
                        slice_dims=(0,))

        self.set_data_reduction_params(data_obj)




from pyreadline.rlmain import Readline

__all__ = [ 'parse_and_bind',
            'get_line_buffer',
            'insert_text',
            'clear_history',
            'read_init_file',
            'read_history_file',
            'write_history_file',
            'get_current_history_length',
            'get_history_length',
            'get_history_item',
            'set_history_length',
            'set_startup_hook',
            'set_pre_input_hook',
            'set_completer',
            'get_completer',
            'get_begidx',
            'get_endidx',
            'set_completer_delims',
            'get_completer_delims',
            'add_history',
            'callback_handler_install',
            'callback_handler_remove',
            'callback_read_char',] #Some other objects are added below


# create a Readline object to contain the state
rl = Readline()

if rl.disable_readline:
    def dummy(completer=""):
        pass
    for funk in __all__:
        globals()[funk] = dummy
else:
    def GetOutputFile():
        '''Return the console object used by readline so that it can be used for printing in color.'''
        return rl.console
    __all__.append("GetOutputFile")

    import pyreadline.console as console

    # make these available so this looks like the python readline module
    read_init_file = rl.read_init_file
    parse_and_bind = rl.parse_and_bind
    clear_history = rl.clear_history
    add_history = rl.add_history
    insert_text = rl.insert_text

    write_history_file = rl.write_history_file
    read_history_file = rl.read_history_file

    get_completer_delims = rl.get_completer_delims
    get_current_history_length = rl.get_current_history_length
    get_history_length = rl.get_history_length
    get_history_item = rl.get_history_item
    get_line_buffer = rl.get_line_buffer
    set_completer = rl.set_completer
    get_completer = rl.get_completer
    get_begidx = rl.get_begidx
    get_endidx = rl.get_endidx

    set_completer_delims = rl.set_completer_delims
    set_history_length = rl.set_history_length
    set_pre_input_hook = rl.set_pre_input_hook
    set_startup_hook = rl.set_startup_hook

    callback_handler_install=rl.callback_handler_install
    callback_handler_remove=rl.callback_handler_remove
    callback_read_char=rl.callback_read_char

    console.install_readline(rl.readline)

__all__.append("rl")


# Copyright 2016 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: threshold_filter
   :platform: Unix
   :synopsis: A plugin to to quantise an image into a number of discrete levels

.. moduleauthor:: Nicoletta De Maio <scientificsoftware@diamond.ac.uk>

"""
import logging
import numpy

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class ThresholdFilter(BaseFilter, CpuPlugin):
    """
    A plugin to quantise an image into discrete levels.

    :param explicit_threshold: False if plugin calculates black/white threshold, True if it's user-defined. Default: True.
    :param intensity_threshold: Threshold for black/white quantisation. Default: 32768.
    """

    def __init__(self):
        """
        This is where the class is designed, and we inherit from 2 classes, the
        first is the Filter class, which deals with splitting the job up for
        us, and provides some simple methods which we need to overload. The
        second is the CpuPlugin class which tells the framework that the
        processing that you are doing here runs on 1 cpu.
        """
        logging.debug("initialising Quantisation filter")
        logging.debug("Calling super to make sure that all superclasses are " +
                      " initialised")
        super(ThresholdFilter,
              self).__init__("ThresholdFilter")
        self.threshold = None
        self.lowest = None
        self.highest = None

    def process_frames(self, data):
        """
        The second method we need to implement from the Filter class and the
        part of the code that actually does all the work. the input here 'data'
        will contain the 3D block of data to process, and we need to return the
        data for the single frame in the middle of this. In this case we return
        the same size data as you had originally.
        """
        data = data[0]
        logging.debug("Data frame recieved for processing of shape %s",
                      str(data.shape))

        result = numpy.zeros(data.shape, dtype=data.dtype)

        result[data < self.threshold] = self.lowest
        result[data >= self.threshold] = self.highest

        return result

    def pre_process(self):
        in_dataset = self.get_in_datasets()[0]
        try:
            # Extract the intensity range from the image meta data
            self.lowest = numpy.amin(in_dataset.meta_data.get('min'))
            self.highest = numpy.amax(in_dataset.meta_data.get('max'))
        except KeyError as k:
            logging.error("Caught KeyError in BinaryQuantisation.setup(): %s", str(k))

            # Use defaults suitable for tomo test data
            logging.warning("Using test data min and max intensities.")
            self.lowest = 0
            self.highest = 31137

        if self.parameters['explicit_threshold']:
            self.threshold = self.parameters['intensity_threshold']
        else:
            self.threshold = (self.highest + self.lowest) / 2.0



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: stitch_data
   :platform: Unix
   :synopsis: A module for stitching together multiple datasets.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import numpy as np

from savu.data.data_structures.data_types.base_type import BaseType


class StitchData(BaseType):
    """ This class is used to combine multiple data objects. """

    def __init__(self, data_obj_list, stack_or_cat, dim, remove=[]):
        self.obj_list = data_obj_list
        self.dtype = data_obj_list[0].data.dtype
        self.stack_or_cat = stack_or_cat
        self.dim = dim
        self.remove = remove
        self.dark_updated = False
        self.flat_updated = False
        super(StitchData, self).__init__()

        self.shape = None
        self._set_shape()
        if self.stack_or_cat == 'stack':
            self.inc = 1
            self._getitem = self._getitem_stack
            self._get_lists = self._get_lists_stack
        else:
            self.inc = self.obj_list[0].get_shape()[self.dim]
            self._getitem = self._getitem_cat
            self._get_lists = self._get_lists_cat

    def clone_data_args(self, args, kwargs, extras):
        args = ['obj_list', 'stack_or_cat', 'dim']
        kwargs['remove'] = 'remove'
        extras = ['shape']
        return args, kwargs, extras

    def __getitem__(self, idx):
        size = [len(np.arange(s.start, s.stop, s.step)) for s in idx]
        obj_list, in_slice_list, out_slice_list = self._get_lists(idx)
        data = np.empty(size)

        for i in range(len(obj_list)):
            data[tuple(out_slice_list[i])] = \
                self._getitem(obj_list[i], in_slice_list[i])
        return data

    def _getitem_stack(self, obj, sl):
        data = obj.data[tuple(sl)]
        for i in np.sort(self.remove)[::-1]:
            data = np.squeeze(data, axis=i)
        return np.expand_dims(data, self.dim)

    def _getitem_cat(self, obj, sl):
        data = obj.data[tuple(sl)]
        for i in np.sort(self.remove)[::-1]:
            data = np.squeeze(data, axis=i)
        return data

    def _get_lists_stack(self, idx):
        entry = idx[self.dim]
        init_vals = np.arange(entry.start, entry.stop, entry.step)
        obj_list = []
        for i in init_vals:
            obj_list.append(self.obj_list[i])

        in_idx = list(idx)
        del in_idx[self.dim]
        in_slice_list = np.tile(in_idx, (len(init_vals), 1))

        out_slice_list = \
            [slice(0, len(np.arange(s.start, s.stop, s.step))) for s in idx]
        out_slice_list = np.tile(out_slice_list, (len(init_vals), 1))

        new_slices = [slice(i, i + 1) for i in range(len(init_vals))]
        out_slice_list[:, self.dim] = new_slices
        return obj_list, in_slice_list, out_slice_list

    def _get_lists_cat(self, idx):
        inc = self.inc
        entry = idx[self.dim]
        init_vals = np.arange(entry.start, entry.stop, entry.step)
        array = init_vals % inc
        index = np.where(np.diff(array) < 0)[0] + 1

        val_list = np.array_split(array, index)
        obj_vals = init_vals[np.append(0, index)] / inc
        active_obj_list = []
        for i in obj_vals:
            active_obj_list.append(self.obj_list[i])

        in_slice_list = self._set_in_slice_list(idx, val_list, entry)
        out_slice_list = self._set_out_slice_list(idx, val_list)
        return active_obj_list, in_slice_list, out_slice_list

    def _set_in_slice_list(self, idx, val_list, entry):
        in_slice_list = np.tile(idx, (len(val_list), 1))
        new_slices = [slice(e[0], e[-1] + 1, entry.step) for e in val_list]
        in_slice_list[:, self.dim] = new_slices
        return in_slice_list

    def _set_out_slice_list(self, idx, val_list):
        out_slice_list = \
            [slice(0, len(np.arange(s.start, s.stop, s.step))) for s in idx]
        out_slice_list = np.tile(out_slice_list, (len(val_list), 1))
        length = np.append(0, np.cumsum([len(v) for v in val_list]))
        if self.stack_or_cat == 'cat':
            new_slices = \
                [slice(length[i - 1], length[i])
                 for i in range(1, len(length))]
        else:
            new_slices = [slice(i, i + 1) for i in range(len(val_list))]
        out_slice_list[:, self.dim] = new_slices
        return out_slice_list

    def get_shape(self):
        return self.shape

    def _set_shape(self):
        nObjs = len(self.obj_list)
        shape = list(self.obj_list[0].data.shape)

        for dim in np.sort(self.remove)[::-1]:
            del shape[dim]

        if self.stack_or_cat == 'cat':
            shape[self.dim] *= nObjs
        else:
            shape.insert(self.dim, nObjs)
        self.shape = tuple(shape)

    def update_dark(self, data):
        self.dark_updated = data

    def update_flat(self, data):
        self.flat_updated = data

    def dark(self):
        if self.dark_updated:
            return self.dark_updated
        if self.stack_or_cat == 'stack':
            return np.vstack(tuple(np.asarray([d.data.dark() for d in self.obj_list])))
        else:
            return np.hstack(tuple(np.asarray([d.data.dark() for d in self.obj_list])))

    def flat(self):
        if self.flat_updated:
            return self.flat_updated
        if self.stack_or_cat == 'stack':
            return np.vstack(tuple(np.asarray([d.data.flat() for d in self.obj_list])))
        else:
            return np.hstack(tuple(np.asarray([d.data.flat() for d in self.obj_list])))

    def dark_mean(self):
        """ Get the averaged dark projection data. """
        if self.stack_or_cat == 'stack':
            return np.vstack(tuple(np.asarray([d.data.dark_mean() for d in self.obj_list])))
        else:
            return np.hstack(tuple(np.asarray([d.data.dark_mean() for d in self.obj_list])))

    def flat_mean(self):
        """ Get the averaged flat projection data. """
        if self.stack_or_cat == 'stack':
            return np.vstack(tuple(np.asarray([d.data.flat_mean() for d in self.obj_list])))
        else:
            return np.hstack(tuple(np.asarray([d.data.flat_mean() for d in self.obj_list])))



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: I22TomoLoader
   :platform: Unix
   :synopsis: A class for loading I22

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.utils import register_plugin
from savu.plugins.loaders.base_loader import BaseLoader
import h5py
import logging
import numpy as np


@register_plugin
class I22TomoLoader(BaseLoader):
    """
    A class to load tomography data for I22 that has been previously processed\
    using dawn
    """

    def __init__(self, name='I22TomoLoader'):
        super(I22TomoLoader, self).__init__(name)

    def setup(self):
        """
        """
        exp = self.exp
        data_obj = exp.create_data_object('in_data', 'tomo')
        data_obj.backing_file = \
            h5py.File(exp.meta_data.get("data_file"), 'r')
        data_obj.data = data_obj.backing_file['entry/result/data']
        data_obj.set_shape(data_obj.data.shape)
        logging.warning('the data as shape %s' % str(data_obj.data.shape))
        data_obj.set_axis_labels('y.units', 'x.units',
                                 'rotation_angle.degrees', 'Q.angstrom^-1')

        data_obj.add_pattern('PROJECTION', core_dims=(1, 0), slice_dims=(2, 3))
        data_obj.add_pattern('SINOGRAM', core_dims=(2, 1), slice_dims=(0, 3))
        data_obj.add_pattern('SPECTRUM', core_dims=(3,), slice_dims=(0, 1, 2))

        mData = data_obj.meta_data
        mData.set("Q", data_obj.backing_file['entry/result/q'][()])
        mData.set("x", np.arange(data_obj.data.shape[1]))
        mData.set("y", np.arange(data_obj.data.shape[0]))
        mData.set("rotation_angle", data_obj.backing_file[
            'entry/result/theta'][()])

        self.set_data_reduction_params(data_obj)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: data
   :platform: Unix
   :synopsis: The Data class dynamically inherits from transport specific data\
   class and holds the data array, along with associated information.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import savu.core.utils as cu
from savu.data.meta_data import MetaData
import savu.data.data_structures.utils as dsu
from savu.data.data_structures.preview import Preview
from savu.data.data_structures.data_create import DataCreate


class Data(DataCreate):
    """The Data class dynamically inherits from transport specific data class
    and holds the data array, along with associated information.
    """

    def __init__(self, name, exp):
        super(Data, self).__init__(name)
        self.meta_data = MetaData()
        self.related = {}
        self.pattern_list = self.__get_available_pattern_list()
        self.data_info = MetaData()
        self.__initialise_data_info(name)
        self._preview = Preview(self)
        self.exp = exp
        self.group_name = None
        self.group = None
        self._plugin_data_obj = None
        self.raw = None
        self.backing_file = None
        self.data = None
        self.next_shape = None
        self.orig_shape = None
        self.previous_pattern = None
        self.transport_data = None

#    def get_data(self, related):
#        return self.related[related].data

    def __initialise_data_info(self, name):
        """ Initialise entries in the data_info meta data.
        """
        self.data_info.set('name', name)
        self.data_info.set('data_patterns', {})
        self.data_info.set('shape', None)
        self.data_info.set('nDims', None)

    def _set_plugin_data(self, plugin_data_obj):
        """ Encapsulate a PluginData object.
        """
        self._plugin_data_obj = plugin_data_obj

    def _clear_plugin_data(self):
        """ Set encapsulated PluginData object to None.
        """
        self._plugin_data_obj = None

    def _get_plugin_data(self):
        """ Get encapsulated PluginData object.
        """
        if self._plugin_data_obj is not None:
            return self._plugin_data_obj
        else:
            raise Exception("There is no PluginData object associated with "
                            "the Data object.")

    def get_preview(self):
        """ Get the Preview instance associated with the data object
        """
        return self._preview

    def _set_transport_data(self, transport):
        """ Import the data transport mechanism

        :returns: instance of data transport
        :rtype: transport_data
        """
        transport_data = "savu.data.transport_data." + transport + \
                         "_transport_data"
        transport_data = cu.import_class(transport_data)
        self.transport_data = transport_data(self)
        self.data_info.set('transport', transport)

    def _get_transport_data(self):
        return self.transport_data

    def __deepcopy__(self, memo):
        """ Copy the data object.
        """
        name = self.data_info.get('name')
        return dsu._deepcopy_data_object(self, Data(name, self.exp))

    def get_data_patterns(self):
        """ Get data patterns associated with this data object.

        :returns: A dictionary of associated patterns.
        :rtype: dict
        """
        return self.data_info.get('data_patterns')

    def _set_previous_pattern(self, pattern):
        self.previous_pattern = pattern

    def get_previous_pattern(self):
        return self.previous_pattern

    def set_shape(self, shape):
        """ Set the dataset shape.
        """
        self.data_info.set('shape', shape)
        self.__check_dims()

    def set_original_shape(self, shape):
        """ Set the original data shape before previewing
        """
        self.orig_shape = shape
        self.set_shape(shape)

    def get_original_shape(self):
        """
        Returns the original shape of the data before previewing
        
        Returns
        -------
        tuple
            Original data shape.
        """
        return self.orig_shape

    def get_shape(self):
        """ Get the dataset shape

        :returns: data shape
        :rtype: tuple
        """
        shape = self.data_info.get('shape')
        return shape

    def __check_dims(self):
        """ Check the ``shape`` and ``nDims`` entries in the data_info
        meta_data dictionary are equal.
        """
        nDims = self.data_info.get("nDims")
        shape = self.data_info.get('shape')
        if nDims:
            if len(shape) != nDims:
                error_msg = ("The number of axis labels, %d, does not "
                             "coincide with the number of data "
                             "dimensions %d." % (nDims, len(shape)))
                raise Exception(error_msg)

    def _set_name(self, name):
        self.data_info.set('name', name)

    def get_name(self, orig=False):
        """ Get data name.

        :keyword bool orig: Set this flag to true to return the original cloned
            dataset name if this dataset is a clone
        :returns: the name associated with the dataset
        :rtype: str
        """
        if orig:
            dinfo = self.data_info.get_dictionary()
            return dinfo['clone'] if 'clone' in dinfo.keys() else dinfo['name']
        return self.data_info.get('name')

    def __get_available_pattern_list(self):
        """ Get a list of ALL pattern names that are currently allowed in the
        framework.
        """
        pattern_list = dsu.get_available_pattern_types()
        return pattern_list

    def add_pattern(self, dtype, **kwargs):
        """ Add a pattern.

        :params str dtype: The *type* of pattern to add, which can be anything
            from the :const:`savu.data.data_structures.utils.pattern_list`
            :const:`pattern_list`
            :data:`savu.data.data_structures.utils.pattern_list`
            :data:`pattern_list`:
        :keyword tuple core_dims: Dimension indices of core dimensions
        :keyword tuple slice_dims: Dimension indices of slice dimensions
        """
        if dtype in self.pattern_list:
            nDims = 0
            for args in kwargs:
                dlen = len(kwargs[args])
                if not dlen:
                    raise Exception("Pattern Error: Pattern %s must have at"
                                    " least one %s" % (dtype, args))
                nDims += len(kwargs[args])
                self.data_info.set(['data_patterns', dtype, args],
                                   kwargs[args])

            self.__convert_pattern_dimensions(dtype)
            if self.get_shape():
                diff = len(self.get_shape()) - nDims
                if diff:
                    pattern = {dtype: self.get_data_patterns()[dtype]}
                    self._set_data_patterns(pattern)
                    nDims += diff
            try:
                if nDims != self.data_info.get("nDims"):
                    actualDims = self.data_info.get('nDims')
                    err_msg = ("The pattern %s has an incorrect number of "
                               "dimensions: %d required but %d specified."
                               % (dtype, actualDims, nDims))
                    raise Exception(err_msg)
            except KeyError:
                self.data_info.set('nDims', nDims)
        else:
            raise Exception("The data pattern '%s'does not exist. Please "
                            "choose from the following list: \n'%s'",
                            dtype, str(self.pattern_list))

    def add_volume_patterns(self, x, y, z):
        """ Adds volume patterns

        :params int x: dimension to be associated with x-axis
        :params int y: dimension to be associated with y-axis
        :params int z: dimension to be associated with z-axis
        """
        self.add_pattern("VOLUME_XZ", **self.__get_dirs_for_volume(x, z, y))

        if y:
            self.add_pattern(
                "VOLUME_YZ", **self.__get_dirs_for_volume(y, z, x))
            self.add_pattern(
                "VOLUME_XY", **self.__get_dirs_for_volume(x, y, z))

        if self.data_info.get("nDims") > 3 and y:
            self.add_pattern("VOLUME_3D", **self.__get_dirs_for_volume_3D())

    def __get_dirs_for_volume(self, dim1, dim2, sdir, dim3=None):
        """ Calculate core_dir and slice_dir for a volume pattern.
        """
        all_dims = range(self.data_info.get("nDims"))
        vol_dict = {}
        vol_dict['core_dims'] = (dim1, dim2)
        slice_dir = [sdir] if type(sdir) is int else []
        for ddir in all_dims:
            if ddir not in [dim1, dim2, sdir]:
                slice_dir.append(ddir)                
        vol_dict['slice_dims'] = tuple(slice_dir)
        return vol_dict

    def __get_dirs_for_volume_3D(self):
        # create volume 3D pattern here
        patterns = self.get_data_patterns()
        cdim = []
        for v in ['VOLUME_YZ', 'VOLUME_XY', 'VOLUME_XZ']:
            cdim += (patterns[v]['core_dims'])
            
        cdim = set(cdim)
        sdim = tuple(set(range(self.data_info.get("nDims"))).difference(cdim))
        return {"core_dims": tuple(cdim), "slice_dims": sdim}

    def set_axis_labels(self, *args):
        """ Set the axis labels associated with each data dimension.

        :arg str: Each arg should be of the form ``name.unit``. If ``name`` is\
        a data_obj.meta_data entry, it will be output to the final .nxs file.
        """
        self.data_info.set('nDims', len(args))
        axis_labels = []
        for arg in args:
            if isinstance(arg, dict):
                axis_labels.append(arg)
            else:
                try:
                    axis = arg.split('.')
                    axis_labels.append({axis[0]: axis[1]})
                except:
                    # data arrives here, but that may be an error
                    pass
        self.data_info.set('axis_labels', axis_labels)

    def get_axis_labels(self):
        """ Get axis labels.

        :returns: Axis labels
        :rtype: list(dict)
        """
        return self.data_info.get('axis_labels')

    def get_data_dimension_by_axis_label(self, name, contains=False, exists=False):
        """ Get the dimension of the data associated with a particular
        axis_label.

        :param str name: The name of the axis_label
        :keyword bool contains: Set this flag to true if the name is only part
            of the axis_label name
        :keyword bool exists: Set to True to return False rather than Exception
        :returns: The associated axis number
        :rtype: int
        """
        axis_labels = self.data_info.get('axis_labels')
        for i in range(len(axis_labels)):
            if contains is True:
                for names in list(axis_labels[i].keys()):
                    if name in names:
                        return i
            else:
                if name in list(axis_labels[i].keys()):
                    return i
        if exists:
            return False
        raise Exception("Cannot find the specifed axis label.")

    def _finalise_patterns(self):
        """ Adds a main axis (fastest changing) to SINOGRAM and PROJECTON
        patterns.
        """
        check = 0        
        check += self.__check_pattern('SINOGRAM')
        check += self.__check_pattern('PROJECTION')

        if check == 2 and len(self.get_shape()) > 2:
            self.__set_main_axis('SINOGRAM')
            self.__set_main_axis('PROJECTION')

    def __check_pattern(self, pattern_name):
        """ Check if a pattern exists.
        """
        patterns = self.get_data_patterns()
        try:
            patterns[pattern_name]
        except KeyError:
            return 0
        return 1

    def __convert_pattern_dimensions(self, dtype):
        """ Replace negative indices in pattern kwargs.
        """
        pattern = self.get_data_patterns()[dtype]
        if 'main_dir' in list(pattern.keys()):
            del pattern['main_dir']

        nDims = sum([len(i) for i in list(pattern.values())])
        for p in pattern:
            ddirs = pattern[p]
            pattern[p] = self._non_negative_directions(ddirs, nDims)

    def _non_negative_directions(self, ddirs, nDims):
        """ Replace negative indexing values with positive counterparts.

        :params tuple(int) ddirs: data dimension indices
        :params int nDims: The number of data dimensions
        :returns: non-negative data dimension indices
        :rtype: tuple(int)
        """
        index = [i for i in range(len(ddirs)) if ddirs[i] < 0]
        list_ddirs = list(ddirs)
        for i in index:
            list_ddirs[i] = nDims + ddirs[i]
        return tuple(list_ddirs)

    def __set_main_axis(self, pname):
        """ Set the ``main_dir`` pattern kwarg to the fastest changing
        dimension
        """
        patterns = self.get_data_patterns()
        n1 = 'PROJECTION' if pname is 'SINOGRAM' else 'SINOGRAM'
        d1 = patterns[n1]['core_dims']
        d2 = patterns[pname]['slice_dims']
        tdir = set(d1).intersection(set(d2))

        # this is required when a single sinogram exists in the mm case, and a
        # dimension is added via parameter tuning.
        if not tdir:
            tdir = [d2[0]]

        self.data_info.set(['data_patterns', pname, 'main_dir'], list(tdir)[0])

    def get_axis_label_keys(self):
        """ Get axis_label names

        :returns: A list containing associated axis names for each dimension
        :rtype: list(str)
        """
        axis_labels = self.data_info.get('axis_labels')
        axis_label_keys = []
        for labels in axis_labels:
            for key in list(labels.keys()):
                axis_label_keys.append(key)
        return axis_label_keys

    def amend_axis_label_values(self, slice_list):
        """ Amend all axis label values based on the slice_list parameter.\
        This is required if the data is reduced.
        """
        axis_labels = self.get_axis_labels()
        for i in range(len(slice_list)):
            label = list(axis_labels[i].keys())[0]
            if label in list(self.meta_data.get_dictionary().keys()):
                values = self.meta_data.get(label)
                preview_sl = [slice(None)]*len(values.shape)
                preview_sl[0] = slice_list[i]
                self.meta_data.set(label, values[tuple(preview_sl)])

    def get_core_dimensions(self):
        """ Get the core data dimensions associated with the current pattern.

        :returns: value associated with pattern key ``core_dims``
        :rtype: tuple
        """
        return list(self._get_plugin_data().get_pattern().values())[0]['core_dims']

    def get_slice_dimensions(self):
        """ Get the slice data dimensions associated with the current pattern.

        :returns: value associated with pattern key ``slice_dims``
        :rtype: tuple
        """
        return list(self._get_plugin_data().get_pattern().values())[0]['slice_dims']

    def get_itemsize(self):
        """ Returns bytes per entry """
        dtype = self.get_dtype()
        if not dtype:
            self.set_dtype(None)
            dtype = self.get_dtype()
        return self.get_dtype().itemsize



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: stxm_analysis
   :platform: Unix
   :synopsis: A plugin to analyse STXM data from area detectors

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""
import logging
from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
import numpy as np
from scipy.ndimage import center_of_mass
@register_plugin
class StxmAnalysis(BaseFilter, CpuPlugin):
    """
    This plugin performs basic STXM analysis of diffraction patterns.
    :param mask_file: takes in a mask currently in hdf format. Default: None.
    :param mask_path: path to the mask inside the file. Default:'/mask'.
    :param threshold: intensity threshold for the dark field. Default: 0.05.
    :param out_datasets: A. Default: ["bf","df","dpc_x","dpc_y","combined_dpc"].
    """

    def __init__(self):
        super(StxmAnalysis, self).__init__("StxmAnalysis")

    def pre_process(self):
        if self.parameters['mask_file'] is not None:
            import h5py as h5
            self.mask = h5.File(self.parameters['mask_file'],'r')[self.parameters['mask_path']]
        else:
            self.mask = None
#         print self.mask.shape


    def filter_frames(self, data):
        d=data[0]
        d = d*self.mask #  set everything to zero that's in the mask
        bf = np.array([d.sum().sum()])
        df= np.array([d[d<self.parameters['threshold']].sum().sum()])
        dpc_x, dpc_y = center_of_mass(d)
        combined_dpc = np.array([np.sqrt(dpc_x**2 + dpc_y**2)])
        return [bf, df, np.array([dpc_x]), np.array([dpc_y]), combined_dpc]

    def setup(self):
        # set up the output datasets that are created by the plugin
        in_dataset, out_datasets = self.get_datasets()
        shape = in_dataset[0].get_shape()
        new_shape = shape[:-2]+(1,) # we will assume diffraction is the last two slices
        axis_labels = ['-2', '-1.Value.au']
        patterns = ['SINOGRAM.-1', 'PROJECTION.-1', '4D_SCAN.-1']
        bf = out_datasets[0]
        df = out_datasets[1]
        dpc_x = out_datasets[2]
        dpc_y = out_datasets[3]
        combined_dpc = out_datasets[4]
        in_dataset[0].data_info.get_data('axis_labels')
        
        bf.create_dataset(patterns={in_dataset[0]: patterns},
                          axis_labels={in_dataset[0]: axis_labels},
                          shape=new_shape)

        df.create_dataset(patterns={in_dataset[0]: patterns},
                          axis_labels={in_dataset[0]: axis_labels},
                          shape=new_shape)

        dpc_x.create_dataset(patterns={in_dataset[0]: patterns},
                             axis_labels={in_dataset[0]: axis_labels},
                             shape=new_shape)

        dpc_y.create_dataset(patterns={in_dataset[0]: patterns},
                             axis_labels={in_dataset[0]: axis_labels},
                             shape=new_shape)

        combined_dpc.create_dataset(patterns={in_dataset[0]: patterns},
                                    axis_labels={in_dataset[0]: axis_labels},
                                    shape=new_shape)

        channel = {'core_dims': (-1,), 'slice_dims': list(range(len(new_shape)-1))}

        bf.add_pattern("CHANNEL", **channel)
        df.add_pattern("CHANNEL", **channel)
        dpc_x.add_pattern("CHANNEL", **channel)
        dpc_y.add_pattern("CHANNEL", **channel)
        combined_dpc.add_pattern("CHANNEL", **channel)

        # setup plugin datasets
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('DIFFRACTION', self.get_max_frames())

        out_pData[0].plugin_data_setup('CHANNEL', self.get_max_frames())
        out_pData[1].plugin_data_setup('CHANNEL', self.get_max_frames())
        out_pData[2].plugin_data_setup('CHANNEL', self.get_max_frames())
        out_pData[3].plugin_data_setup('CHANNEL', self.get_max_frames())
        out_pData[4].plugin_data_setup('CHANNEL', self.get_max_frames())

    def nOutput_datasets(self):
        return 5

    def get_max_frames(self):
        return 'single'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: chunking
   :platform: Unix
   :synopsis: A class to optimise hdf5 chunking
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""

import copy
import logging
from math import gcd
import numpy as np


class Chunking(object):
    """
    A class to save tomography data to a hdf5 file
    """

    def __init__(self, exp, patternDict):
        self.pattern_dict = patternDict
        self.current = patternDict['current'][list(patternDict['current'].keys())[0]]
        if patternDict['next']:
            self.next = patternDict['next'][list(patternDict['next'].keys())[0]]
        else:
            self.next = self.current

        try:
            self.next_pattern = list(patternDict['next'].keys())[0]
        except AttributeError:
            self.next_pattern = list(patternDict['current'].keys())[0]

        self.exp = exp
        self.core = None
        self.slice1 = None
        self.other = None
        self.default_chunk_max = 1000000

    def __lustre_workaround(self, chunks, shape):
        nChunks_to_create_file = \
            np.ceil(np.prod(np.array(shape)/np.array(chunks, dtype=np.float)))
        nProcesses = self.exp.meta_data.get('processes')
        dims = list(range(len(shape)))
        chunks = list(chunks)
        if nChunks_to_create_file < nProcesses:
            idx = [i for i in dims if shape[i] - chunks[i] > 0 and
                   chunks[i] > 1]
            idx = idx if idx else [i for i in dims if chunks[i] > 1]
            if idx:
                chunks[idx[0]] = int(np.ceil(chunks[idx[0]] / 2.0))
                return tuple(chunks)
            else:
                raise Exception('There is an error in the lustre workaround')

    def _calculate_chunking(self, shape, ttype, chunk_max=None):
        """
        Calculate appropriate chunk sizes for this dataset
        """
        self.chunk_max = chunk_max if chunk_max else self.default_chunk_max
        logging.debug("shape = %s", shape)
        if len(shape) < 3:
            return True

        chunks = [1]*len(shape)
        adjust = self.__set_adjust_params(shape)
        self.__set_chunks(chunks, shape, adjust)

        if 0 in chunks:
            return True
        else:
            chunks = self.__adjust_chunk_size(chunks, ttype, shape, adjust)
            # temporary work around for lustre
            if self.exp.meta_data.get('lustre') is True:
                chunks = self.__lustre_workaround(chunks, shape)

            logging.debug("chunk size %s", chunks)
            return tuple(chunks)

    def __set_adjust_params(self, shape):
        """
        Set adjustable dimension parameters (the dimension number, increment
        and max value)
        """
        adjust_dim = self.__get_adjustable_dims()
        array_len = len(adjust_dim)
        adjust_max = [1]*array_len

        for i in range(array_len):
            adjust_max[i] = shape[adjust_dim[i]]

        inc_dict = {'up': [1]*array_len, 'down': [1]*array_len}
        bounds = {'min': [1]*array_len, 'max': adjust_max}
        return {'dim': adjust_dim, 'inc': inc_dict, 'bounds': bounds}

    def __get_adjustable_dims(self):
        """
        Get all core dimensions and fastest changing slice dimension (all
        potentially adjustable)
        """
        self.core = \
            list(self.current['core_dims']) + list(self.next['core_dims'])
        c_sl = list(self.current['slice_dims'])
        n_sl = list(self.next['slice_dims'])
        self.slice1 = [c_sl[0]] + [n_sl[0]]
        self.other = c_sl[1:] + n_sl[1:]
        return list(set(self.core + self.slice1))

    def __set_chunks(self, chunks, shape, adjust):
        """
        Calculate initial chunk values
        """
        chunk_functions = {0: self.__slice_slice, 1: self.__slice_other,
                           2: self.__core_core, 3: self.__core_other,
                           4: self.__core_slice}
        adj_idx = 0
        for dim in adjust['dim']:
            count = (dim in self.core)*4 + (dim in self.slice1)*2 + \
                (dim in self.other) - 2
            chunks[dim] = chunk_functions[count](dim, adj_idx, adjust, shape)
#            if 'VOLUME' in self.next_pattern:
#                self.__set_volume_bounds(adjust, dim, chunks)
            adj_idx += 1
        return chunks

    def __set_volume_bounds(self, adjust, dim, chunks):
        adjust['bounds']['min'][dim] = \
            eval(str(62) + adjust['inc']['down'][dim])
        chunks[dim] = int(min(adjust['bounds']['max'][dim], 62))

    def __core_core(self, dim, adj_idx, adjust, shape):
        adjust['inc']['up'][adj_idx] = '+1'
        adjust['inc']['down'][adj_idx] = '/2'
        adjust['bounds']['max'][adj_idx] = shape[dim]
        return shape[dim]

    def __core_slice(self, dim, adj_idx, adjust, shape):
        max_frames = self.__get_max_frames_dict()[dim]
        adjust['inc']['up'][adj_idx] = '+' + str(max_frames)
        adjust['inc']['down'][adj_idx] = '/2'  # '-' + str(max_frames)

        # which is the slice dimension: current or next?
        ddict = self.current if dim in self.current['slice_dims'] else self.next
        shape, allslices = self.__get_shape(shape, ddict)

        adjust['bounds']['max'][adj_idx] = self.__max_frames_per_process(
                shape, max_frames, allslices=allslices)
        return min(max_frames, shape)

    def __core_other(self, dim, adj_idx, adjust, shape):
        adjust['inc']['up'][adj_idx] = '+1'
        adjust['inc']['down'][adj_idx] = '-1'
        adjust['bounds']['max'][adj_idx] = shape[dim]
        return 1

    def __slice_slice(self, dim, adj_idx, adjust, shape):
        max_frames = self.__get_max_frames_dict()[dim]
        adjust['inc']['up'][adj_idx] = '+' + str(max_frames)
        adjust['inc']['down'][adj_idx] = '/2'

        shape1 = np.prod([shape[s] for s in self.current['slice_dims']])
        shape2 = np.prod([shape[s] for s in self.next['slice_dims']])
        ddict = self.current if shape1 < shape2 else self.next
        shape, allslices = self.__get_shape(shape, ddict)
        adjust['bounds']['max'][adj_idx] = self.__max_frames_per_process(
                shape, max_frames, allslices=allslices)
        return min(max_frames, shape)

    def __slice_other(self, dim, adj_idx, adjust, shape):
        adjust['inc']['up'][adj_idx] = '+1'
        adjust['inc']['down'][adj_idx] = '-1'
        adjust['bounds']['max'][adj_idx] = shape[dim]
        return 1

    def __get_max_frames_dict(self):
        current_sdir = self.current['slice_dims'][0]
        next_sdir = self.next['slice_dims'][0]
        mft = 'max_frames_transfer'
        if current_sdir == next_sdir:
            c_max = self.current[mft]
            n_max = self.next[mft]
            least_common_multiple = (c_max*n_max) // gcd(c_max, n_max)
            ddict = {current_sdir: least_common_multiple}
        else:
            ddict = {self.current['slice_dims'][0]: self.current[mft],
                     self.next['slice_dims'][0]: self.next[mft]}
        return ddict

    def __get_shape(self, shape, ddict):
        """ Get shape taking into account padding. """
        shape = [shape[s] for s in ddict['slice_dims']]
        if 'transfer_shape' not in list(ddict.keys()):
            return shape[0], np.prod(shape)
        size_list = [ddict['transfer_shape'][s] for s in ddict['slice_dims']]
        trans_per_dim = np.ceil(np.array(shape)/np.array(
                size_list, dtype=np.float32))
        full_shape = np.prod(trans_per_dim*size_list)
        return shape[0], full_shape

    def __max_frames_per_process(self, shape, nFrames, allslices=None):
        """
        Calculate the max possible frames per process
        """
        nSlices = allslices if allslices else shape
        total_plugin_runs = np.ceil(float(nSlices) / nFrames)
        frame_list = np.arange(total_plugin_runs)
        nProcs = len(self.exp.meta_data.get('processes'))
        frame_list_per_proc = np.array_split(frame_list, nProcs)
        flist_len = [len(f) for f in frame_list_per_proc if len(f) > 0]
        runs_per_proc = int(np.median(np.array(flist_len)))
        return int(min(runs_per_proc*nFrames, shape))

    def __adjust_chunk_size(self, chunks, ttype, shape, adjust):
        """
        Adjust the chunk size to as close to 1MB as possible
        """
        chunks = np.array(chunks)
        chunk_size = np.prod(chunks)*np.dtype(ttype).itemsize
        cache_size = self.chunk_max
        if (chunk_size > cache_size):
            self.__decrease_chunks(chunks, ttype, adjust)
        else:
            chunks = self.__increase_chunks(chunks, ttype, shape, adjust)
        return tuple(chunks)

    def __decrease_chunks(self, chunks, ttype, adjust):
        """
        Decrease the chunk size to below but as close to 1MB as possible
        """
        while ((np.prod(chunks)*np.dtype(ttype).itemsize) > self.chunk_max):
            idx = self.__get_idx_decrease(chunks, adjust)
            dim = adjust['dim'].index(idx)
#            if idx == -1:
#                break
            chunks[idx] = int(np.ceil(eval(str(float(chunks[idx])) +
                              adjust['inc']['down'][dim])))

    def __increase_chunks(self, chunks, ttype, shape, adjust):
        """
        Increase the chunk size as close to 1MB as possible
        """
        next_chunks = copy.copy(chunks)
        while ((np.prod(next_chunks)*np.dtype(ttype).itemsize) <=
                self.chunk_max):
            chunks = copy.copy(next_chunks)
            idx = self.__get_idx_increase(next_chunks, adjust)
            if idx == -1:
                break
            dim = adjust['dim'].index(idx)
            next_chunks[idx] = \
                eval(str(next_chunks[idx]) + adjust['inc']['up'][dim])
        return chunks

    def __get_idx_decrease(self, chunks, adjust):
        """
        Determine the chunk dimension to decrease
        """
        self.check = lambda a, b, c, i: \
            True if (eval(str(a) + b[i])) < c['min'][i] else False
        self.__check_adjust_dims(adjust, chunks, 'down')
        return self.__get_idx_order(adjust, chunks, 'down')

    def __get_idx_increase(self, chunks, adjust):
        """
        Determine the chunk dimension to increase
        """
        self.check = lambda a, b, c, i: \
            True if (eval(str(a) + b[i])) > c['max'][i] else False
        self.__check_adjust_dims(adjust, chunks, 'up')
        return self.__get_idx_order(adjust, chunks, 'up')

    def __get_idx_order(self, adjust, chunks, direction):
        process_order = [self.slice1, self.core]
        sl = slice(None, None, -1)
        if direction is 'up':
            sl = slice(None, None, 1)
            process_order = process_order[::-1]

        avail = list(set(adjust['dim']).intersection(process_order[0]))
        idx_order = np.argsort(chunks[avail])[sl]

        if not idx_order.size:
            avail = list(set(adjust['dim']).intersection(process_order[1]))
            idx_order = np.argsort(chunks[avail])[sl]

        return avail[idx_order[0]] if idx_order.size else -1

    def __check_adjust_dims(self, adjust, chunks, up_down):
        nDel = 0
        for i in range(len(adjust['dim'])):
            i -= nDel
            dim = adjust['dim'][i]
#            if self.check(chunks[dim], adjust['inc'][up_down],
#                          adjust['bounds'], i):
#                adjust['inc']['down'][i] = '-1'
            if self.check(chunks[dim], adjust['inc'][up_down],
                          adjust['bounds'], i):
                del adjust['dim'][i]
                del adjust['inc']['up'][i]
                del adjust['inc']['down'][i]
                del adjust['bounds']['max'][i]
                del adjust['bounds']['min'][i]
                nDel += 1
